/*==============================================================================
  
   Purpose:    low level interface
   Author:     Rudolf Berrendorf
               Computer Science Department
               University of Applied Sciences Bonn-Rhein-Sieg
	       53754 Sankt Augustin, Germany
               rudolf.berrendorf@fh-bonn-rhein-sieg.de
   Date:       20.10.2002
  
==============================================================================*/


#include <pcl.h>

void do_work(){}

int main(int argc, char **argv)
{
    int res, counter_index;
    PCL_BIT_MASK_TYPE counter_used_mask;
    PCL_DRIVER_COMMAND_TYPE commands[PCL_COUNTER_MAX];
    unsigned int count_mode;
    PCL_CNT_TYPE counter_values[PCL_COUNTER_MAX];
    PCL_DESCR_TYPE descr;
    PCL_PROCESSOR_INFO info;


    /* Allocate a handle */
    if(PCLinit(&descr) != PCL_SUCCESS)
        printf("cannot get handle\n");

    /* get processor info */
    if((res = PCL_driver_info(&info)) != PCL_SUCCESS)
      printf("error on PCL_driver_info (%d)\n", res);
    else
      {
	printf("processor vendor  : %s\n", info.vendor);
	printf("processor family  : %s\n", info.family);
	printf("processor model   : %s\n", info.model);
	printf("processor speed   : %d MHz\n", info.mhz);
	printf("number of counters: >=%d\n", info.ncounters);
      }

    /* open driver */
    if((res = PCL_driver_open(descr)) != PCL_SUCCESS)
      printf("error on PCL_driver_open (%d)\n", res);

    /* This is for a DEC Alpha 21164: count FP operations */
    counter_index = 1;
    commands[counter_index] = 0x0a;

    counter_used_mask = (0x1 << counter_index);
    count_mode = PCL_MODE_USER;
    if((res = PCL_driver_start(descr, PCL_COUNTER_MAX, counter_used_mask,
			       commands, count_mode)) != PCL_SUCCESS)
      printf("error on PCL_driver_start (%d)\n", res);

    /* do some work */
    do_work();

    /* read counter */
    if((res = PCL_driver_read(descr, PCL_COUNTER_MAX, counter_used_mask,
			      counter_values)) != PCL_SUCCESS)
      printf("error on PCL_driver_read (%d)\n", res);
    else
      printf("%.0f floating point instructions\n",
	     (double)counter_values[counter_index]);

    /* stop counting */
    if((res = PCL_driver_stop(descr)) != PCL_SUCCESS)
      printf("error on PCL_driver_stop (%d)\n", res);

    /* close driver */
    if((res = PCL_driver_close(descr)) != PCL_SUCCESS)
      printf("error on PCL_driver_close (%d)\n", res);

    /* Deallocate handle */
    if(PCLexit(descr) != PCL_SUCCESS)
        printf("cannot release handle\n");

    return 0;
}

/*============================================================================*
 *                             that's all folks                               *
 *============================================================================*/
