/*==============================================================================


   Purpose:    low level routines for Alpha/Compaq


   Author:     Rudolf Berrendorf
               University of Applied Sciences Bonn-Rhein-Sieg
               Computer Science Department
               53754 Sankt Augustin
               rudolf.berrendorf@fh-bonn-rhein-sieg.de


   last mod:   19.10.2002


   DISCLAIMER

                  PCL - The Performance Counter Library


   Copyright (C) 2000
   Forschungszentrum Juelich GmbH, Federal Republic of Germany.
   Copyright (C) 2001,2002
   Rudolf Berrendorf
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are met:

   Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

     - Redistributions of source code must retain the above copyright notice,
       this list of conditions and the following disclaimer.

     - Redistributions in binary form must reproduce the above copyright
       notice, this list of conditions and the following disclaimer in the
       documentation and/or other materials provided with the distribution.

     - Any publications that result from the use of this software shall
       reasonably refer to the Research Centre's development.

     - All advertising materials mentioning features or use of this software
       must display the following acknowledgement:

           This product includes software developed by Forschungszentrum
           Juelich GmbH, Federal Republic of Germany.

     - Forschungszentrum Juelich GmbH is not obligated to provide the user with
       any support, consulting, training or assistance of any kind with regard
       to the use, operation and performance of this software or to provide
       the user with any updates, revisions or new versions.


   THIS SOFTWARE IS PROVIDED BY FORSCHUNGSZENTRUM JUELICH GMBH "AS IS" AND ANY
   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
   DISCLAIMED. IN NO EVENT SHALL FORSCHUNGSZENTRUM JUELICH GMBH BE LIABLE FOR
   ANY SPECIAL, DIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
   RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
   CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
   CONNECTION WITH THE ACCESS, USE OR PERFORMANCE OF THIS SOFTWARE.
  
==============================================================================*/

#include "../../defines.h"
#include "../../pcl.h"
#include "hw_driver_defs.h"
#include "../../pcl_hw_driver.h"
#include "../../hw_driver_global_defs.h"
#include "../../state_descr.h"
#include "../../determine_mhz.h"

#include <stdlib.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/pfcntr.h>
#include <sys/ioctl.h>
#include <stropts.h>
#include <unistd.h>
#include <sys/processor.h>
#include <sys/resource.h>
#include <sys/sysinfo.h>
#include <machine/hal_sysinfo.h>

#include "read_cycle_counter.h"
#include "cpu_version.h"

/*                                                                            */
/*----------------------------------------------------------------------------*/
/* local variables                                                            */

/* invalid entry-data if event is not available */
#define PCL_INVALID    0, {{{-1,-1,-1,-1},{0,0,0,0}}}

/* entries are control codes for hardware performance counter */
static const PCL_event_descr_t event_table_21164[PCL_MAX_EVENT] =
{
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L1CACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L1CACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L1CACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L1CACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L1CACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L1DCACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L1DCACHE_WRITE */
  {PCL_AVAIL,NOT_COMP,{{{-1,PF5_MUX1_C_DCACHE_ACCESS,-1,-1},{-1,OP_ADD,-1,-1}}}}, /* PCL_L1DCACHE_READWRITE */
  {PCL_AVAIL,COMPOSED,{{{-1,PF5_MUX1_C_DCACHE_ACCESS,PF5_MUX2_C_DCACHE_MISSES,-1},{-1,OP_ADD,OP_SUB,-1}}}}, /* PCL_L1DCACHE_HIT */
  {PCL_AVAIL,NOT_COMP,{{{-1,-1,PF5_MUX2_C_DCACHE_MISSES,-1},{-1,-1,OP_ADD,-1}}}},	/* PCL_L1DCACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L1ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L1ICACHE_WRITE */
  {PCL_AVAIL,NOT_COMP,{{{-1,PF5_MUX1_C_ICACHE_ACCESS,-1,-1},{-1,OP_ADD,-1,-1}}}}, /* PCL_L1ICACHE_READWRITE */
  {PCL_AVAIL,COMPOSED,{{{-1,PF5_MUX1_C_ICACHE_ACCESS,PF5_MUX2_C_ICACHE_MISSES,-1},{-1,OP_ADD,OP_SUB,-1}}}}, /* PCL_L1ICACHE_HIT */
  {PCL_AVAIL,NOT_COMP,{{{-1,-1,PF5_MUX2_C_ICACHE_MISSES,-1},{-1,-1,OP_ADD,-1}}}},	/* PCL_L1ICACHE_MISS */
  {PCL_AVAIL,NOT_COMP,{{{-1,PF5_MUX1_C_SCACHE_READ,-1,-1},{-1,OP_ADD,-1,-1}}}}, /* PCL_L2CACHE_READ */
  {PCL_AVAIL,NOT_COMP,{{{-1,PF5_MUX1_C_SCACHE_WRITE,-1,-1},{-1,OP_ADD,-1,-1}}}}, /* PCL_L2CACHE_WRITE */
  {PCL_AVAIL,NOT_COMP,{{{-1,PF5_MUX1_C_SCACHE_ACCESS,-1,-1},{-1,OP_ADD,-1,-1}}}}, /* PCL_L2CACHE_READWRITE */
  {PCL_AVAIL,COMPOSED,{{{-1,PF5_MUX1_C_SCACHE_ACCESS,PF5_MUX2_C_SCACHE_MISS,-1},{-1,OP_ADD,OP_SUB,-1}}}}, /* PCL_L2CACHE_HIT */
  {PCL_AVAIL,NOT_COMP,{{{-1,-1,PF5_MUX2_C_SCACHE_MISS,-1},{-1,-1,OP_ADD,-1}}}},	/* PCL_L2CACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L2DCACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L2DCACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L2DCACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L2DCACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L2DCACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L2ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L2ICACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L2ICACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L2ICACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L2ICACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_TLB_HIT */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_TLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_ITLB_HIT */
  {PCL_AVAIL,NOT_COMP,{{{-1,-1,PF5_MUX2_C_ITB_MISS,-1},{-1,-1,OP_ADD,-1}}}}, /* PCL_ITLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_DTLB_HIT */
  {PCL_AVAIL,NOT_COMP,{{{-1,-1,PF5_MUX2_C_DTB_MISS,-1},{-1,-1,OP_ADD,-1}}}}, /* PCL_DTLB_MISS */
  {PCL_AVAIL,NOT_COMP,{{{PF5_MUX0_CYCLES,-1,PF5_MUX2_C_CYCLES,-1},{OP_ADD,-1,OP_ADD,-1}}}}, /* PCL_CYCLES */
  /* special case with dummy entry */
  {PCL_AVAIL,NOT_COMP,{{{-1,-1,-1,0x00},{-1,-1,-1,OP_ADD}}}}, /* PCL_ELAPSED_CYCLES */
  {PCL_AVAIL,NOT_COMP,{{{-1,PF5_MUX1_INTEGER_OPERATE,-1,-1},{-1,OP_ADD,-1,-1}}}}, /* PCL_INTEGER_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{-1,PF5_MUX1_C_FP_INSTRUCTIONS,-1,-1},{-1,OP_ADD,-1,-1}}}}, /* PCL_FP_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{-1,PF5_MUX1_C_LOAD_INSTRUCTIONS,-1,-1},{-1,OP_ADD,-1,-1}}}}, /* PCL_LOAD_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{-1,PF5_MUX1_C_STORE_INSTRUCTIONS,-1,-1},{-1,OP_ADD,-1,-1}}}}, /* PCL_STORE_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_LOADSTORE_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{PF5_MUX0_ISSUES,-1,-1,-1},{OP_ADD,-1,-1,-1}}}}, /* PCL_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_JUMP_SUCCESS */
  {PCL_AVAIL,NOT_COMP,{{{-1,-1,PF5_MUX2_C_BRANCH_MISPREDICTS,-1},{-1,-1,OP_ADD,-1}}}}, /* PCL_JUMP_UNSUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_JUMP */
  {PCL_AVAIL,NOT_COMP,{{{-1,-1,PF5_MUX2_C_LOAD_LOCKED,-1},{-1,-1,OP_ADD,-1}}}}, /* PCL_ATOMIC_SUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_ATOMIC_UNSUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_ATOMIC */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_STALL_INTEGER */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_STALL_FP */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_STALL_JUMP */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_STALL_LOAD */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_STALL_STORE */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_STALL */
  {PCL_AVAIL,COMPOSED,{{{-1,PF5_MUX1_C_FP_INSTRUCTIONS,PF5_MUX2_C_CYCLES,-1},{-1,OP_FP_2|OP_FP_STORE,OP_FP_2|OP_FP_DIV_MHZ,-1}}}}, /* PCL_MFLOPS */
  {PCL_AVAIL,COMPOSED,{{{PF5_MUX0_ISSUES,-1,PF5_MUX2_C_CYCLES,-1},{OP_FP_2|OP_FP_STORE,-1,OP_FP_2|OP_FP_DIV,-1}}}}, /* PCL_IPC */
  {PCL_AVAIL,COMPOSED,{{{-1,PF5_MUX1_C_DCACHE_ACCESS,PF5_MUX2_C_DCACHE_MISSES,-1},{-1,OP_FP_1|OP_FP_2|OP_FP_REZ,OP_FP_2|OP_FP_MUL,-1}}}}, /* PCL_L1DCACHE_MISSRATE */
  {PCL_AVAIL,COMPOSED,{{{-1,PF5_MUX1_C_SCACHE_ACCESS,PF5_MUX2_C_SCACHE_MISS,-1},{-1,OP_FP_1|OP_FP_2|OP_FP_REZ,OP_FP_2|OP_FP_MUL,-1}}}}, /* PCL_L2DCACHE_MISSRATE */
  {PCL_UNAVAIL, PCL_INVALID}                     /* PCL_MEM_FP_RATIO */
};


/* logical constants for 21264 systems are available in <sys/pfcntr.h> */
/* entries are control codes for hardware performance counter */
#if !defined(OS_HAS_SUPPORT_FOR_21264)
/* older versions of Digital Unix have no support for 21264 */
static const PCL_event_descr_t event_table_21264[PCL_MAX_EVENT];
#else
static const PCL_event_descr_t event_table_21264[PCL_MAX_EVENT] =
{
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L1CACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L1CACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L1CACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L1CACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L1CACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L1DCACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L1DCACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L1DCACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L1DCACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	                  /* PCL_L1DCACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L1ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L1ICACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L1ICACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L1ICACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	                  /* PCL_L1ICACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L2CACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L2CACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L2CACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L2CACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	                  /* PCL_L2CACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L2DCACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L2DCACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L2DCACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L2DCACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L2DCACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L2ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L2ICACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L2ICACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L2ICACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L2ICACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_TLB_HIT */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_TLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_ITLB_HIT */
  {PCL_AVAIL,NOT_COMP,{{{-1,PF6_MUX1_RET_ITB_MISSES,-1,-1},{-1,OP_ADD,-1,-1}}}}, /* PCL_ITLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_DTLB_HIT */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_DTLB_MISS */
  {PCL_AVAIL,NOT_COMP,{{{PF6_MUX0_CYCLES,PF6_MUX1_CYCLES,-1,-1},{OP_ADD,OP_ADD,-1,-1}}}}, /* PCL_CYCLES */
  /* special case with dummy entry */
  {PCL_AVAIL,NOT_COMP,{{{-1,-1,0x00,-1},{-1,-1,OP_ADD,-1}}}}, /* PCL_ELAPSED_CYCLES */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_INTEGER_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_FP_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_LOAD_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_STORE_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_LOADSTORE_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{PF6_MUX0_RET_INSTRUCTIONS,-1,-1,-1},{OP_ADD,-1,-1,-1}}}},  /* PCL_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_JUMP_SUCCESS */
#if 0
  /* according to the documentation of pfm the event for mispredicted branches
     does actually count the number of cycles.
     */
  {PCL_AVAIL,NOT_COMP,{{{-1,PF6_MUX1_RET_BRANCH_MISPREDICTS,-1,-1},{-1,OP_ADD,-1,-1}}}}, /* PCL_JUMP_UNSUCCESS */
#else
  {PCL_UNAVAIL ,PCL_INVALID},                     /* PCL_JUMP_UNSUCCESS */
#endif
  {PCL_AVAIL,NOT_COMP,{{{-1,PF6_MUX1_RET_COND_BRANCHES,-1,-1},{-1,OP_ADD,-1,-1}}}}, /* PCL_JUMP */
  {PCL_UNAVAIL ,PCL_INVALID},                     /* PCL_ATOMIC_SUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_ATOMIC_UNSUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_ATOMIC */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_STALL_INTEGER */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_STALL_FP */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_STALL_JUMP */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_STALL_LOAD */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_STALL_STORE */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_STALL */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_MFLOPS */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_IPC */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L1DCACHE_MISSRATE */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L2DCACHE_MISSRATE */
  {PCL_UNAVAIL, PCL_INVALID}                      /* PCL_MEM_FP_RATIO */
};
#endif

static int *cpu2fd_mapping = NULL;

/*                                                                            */
/*============================================================================*/
/* get information on processor                                               */

int
PCL_driver_info(
                 PCL_PROCESSOR_INFO *info  /* I: address of info struct */
               )

{
  info->vendor = "DEC";

  /* get processor type */
  cpu = cpu_implementation_version();
  if(cpu == 1)
    info->family = "21164";
  else if(cpu == 2)
    info->family = "21264";
  else
    /* something we can't handle yet */
    info->family = "unknown";
  info->model = "";
  info->mhz = PCL_determine_mhz_rate();
  info->ncounters = 4;

  return PCL_SUCCESS;
}

/*                                                                            */
/*============================================================================*/
/*                                                                            */


#if defined(OS_HAS_SUPPORT_FOR_21264)
/* the definition of struct pfcntrs in <sys/pfcntr.h> for the EV6
   (which should be the same as for the EV4 as claimed by the docs)
   doesn't match the way data is stored by the pfm-driver. The definition
   in the include file uses the sequence <cntr0,cntr1,cycle> while the
   driver uses the following definition (which we have to use, too):
*/

struct pfcntrs_ev6
{
  ulong pr_cycle;    /* process cycle counter */
  ulong pf_cntr0;    /* driver's counter 1 */
  ulong pf_cntr1;    /* driver's counter 2 */
};

#endif


/*                                                                            */
/*============================================================================*/
/* We determine what processor type the program is running on and therefore
   what events are supported. 
   Sets "PCL_event_table" accordingly to get the right register settings.
   Additionally, the pseudo device is opened and the process gets bound
   to the processor.
   Returns PCL_SUCCESS if processor type is supported.
*/

int
PCL_driver_open(
		PCL_DESCR_TYPE d /* I: descriptor */
		)

{
  long cpu;
  struct cpu_info cpu_info;
  long max_cpu;
  PCL_state_descr *descr;
  HW_state_descr *hw_descr;
  int i;

  /* get state descriptor */
  descr = d;
  hw_descr = &descr->hw_state_descr;

  /* allowed count mode */
  descr->allowed_count_mode = PCL_MODE_USER_SYSTEM;

  /* get processor type */
  cpu = cpu_implementation_version();
  if(cpu == 1)
    descr->event_table = event_table_21164;
  else if(cpu == 2)
    descr->event_table = event_table_21264;
  else
    /* something we can't handle yet */
    return PCL_NOT_SUPPORTED;

  /* determine highest available processor number + 1 */
  if(getsysinfo(GSI_CPU_INFO, (caddr_t)&cpu_info, sizeof(cpu_info),
                NULL, NULL, NULL) < 1)
    return PCL_FAILURE;
  else
    max_cpu = cpu_info.ncpus;

  /* allocate dynamic data structutes */
  if((hw_descr->cntrs_ev5 = (struct pfcntrs_ev5 *)
      calloc(max_cpu, sizeof(*(hw_descr->cntrs_ev5)))) == NULL)
    return PCL_FAILURE;
#if defined(OS_HAS_SUPPORT_FOR_21264)
  if((hw_descr->cntrs_ev6 = (struct pfcntrs_ev6 *)
      calloc(max_cpu, sizeof(*(hw_descr->cntrs_ev6)))) == NULL)
    return PCL_FAILURE;
#endif
  if(!cpu2fd_mapping) {
   if((cpu2fd_mapping = malloc(max_cpu * sizeof(int))) == NULL)
    return PCL_FAILURE;
   
   for(i = 0; i < max_cpu;i++)
    cpu2fd_mapping[i] = -1;               
  }

  /* get actual CPU we are running on */
  if(getsysinfo(GSI_CURRENT_CPU, (caddr_t)&hw_descr->cpu_num,
                sizeof(hw_descr->cpu_num), NULL, NULL, NULL) < 1)
   return PCL_FAILURE;

  /* Try to open a channel to the pfcntr driver (if not already opened) */
  if(cpu2fd_mapping[hw_descr->cpu_num] == -1)
    {
      unsigned long cpu_mask;
        
      /* bind process to processor */
      cpu_mask = 1 << hw_descr->cpu_num;
      if (bind_to_cpu(getpid(), cpu_mask, BIND_NO_INHERIT) == -1)
        return PCL_FAILURE;
      
      /* open device (only for this processor) */
      cpu2fd_mapping[hw_descr->cpu_num] = open("/dev/pfcntr", 
                                               O_RDONLY | PCNTOPENONE,
                                               S_IRUSR | S_IRGRP | S_IROTH );

      if (cpu2fd_mapping[hw_descr->cpu_num] == -1) 
        return PCL_FAILURE;       
    }
    
   hw_descr->fd = cpu2fd_mapping[hw_descr->cpu_num]; 

  /* everything ok */
  return PCL_SUCCESS;
}


/*                                                                            */
/*============================================================================*/
/* Starts hardware counters.                                                  */


int
PCL_driver_start(
		 PCL_DESCR_TYPE d, /* I: descriptor */
		 int max_counter,
		 PCL_BIT_MASK_TYPE counter_used_mask,
		 PCL_DRIVER_COMMAND_TYPE *commands,
		 unsigned int count_mode
		 )

{
  long arg, rc, event_select, int_frequency;
  union pmctrs_ev5 pm_ev5;
#if defined(OS_HAS_SUPPORT_FOR_21264)
  union pmctrs_ev6 pm_ev6;
#endif
  PCL_BIT_MASK_TYPE used_mask;
  PCL_state_descr *descr;
  HW_state_descr *hw_descr;


  /* get state descriptor */
  descr = d;
  hw_descr = &descr->hw_state_descr;

  /* set up used mask */
  used_mask = counter_used_mask & PCL_mask[max_counter-1];

  if(descr->event_table == event_table_21164)
    {
      /* 21164 */
      /* initialize selectors */
      event_select = hw_descr->counter_select = int_frequency = 0;
      
      /* extract selected events */
      if(IS_USED(used_mask, 0))
	{
	  event_select = commands[0];
	  hw_descr->counter_select = PF5_SEL_COUNTER_0;
	  int_frequency = PF5_C0_INT_EVERY_65536;
	}
      if(IS_USED(used_mask, 1))
	{
	  event_select |= commands[1];
	  hw_descr->counter_select |= PF5_SEL_COUNTER_1;
	  int_frequency |= PF5_C1_INT_EVERY_65536;
	}
      if(IS_USED(used_mask, 2))
	{
	  event_select |= commands[2];
	  hw_descr->counter_select |= PF5_SEL_COUNTER_2;
	  int_frequency |= PF5_C2_INT_EVERY_16384;
	}

      /* select items (= counters) */
      arg = PFM_COUNTERS;
      rc = ioctl(hw_descr->fd, PCNTSETITEMS, &arg);
      if (rc < 0)
	return PCL_FAILURE;

      /* clear driver's counters */
      rc = ioctl(hw_descr->fd, PCNTCLEARCNT);
      if (rc < 0)
	return PCL_FAILURE;

      /* set contexts in which to count */
      arg = PF5_CTXT_THIS_PROC;
      /* can anybody explain to me why the constants PF5_CTXT_... have a
	 meaning different to that what their name suggests?
	 If you want an explanation for the settings, check how it is done on
	 the CRAY.
      */
      if(count_mode == PCL_MODE_USER)
	/* user mode only: 0x40000200 */
	arg |= PF5_CTXT_USER_AND_KERNEL;
      else if(count_mode == PCL_MODE_SYSTEM)
	/* kernel mode only: 0x40000100 */
	arg |= PF5_CTXT_PAL;
      else
	/* user or kernel mode: 0x00000000 */
	arg |= PF5_CTXT_ALL_MODES;
      rc = ioctl(hw_descr->fd, PCNT5CTXTS, &arg);
      if (rc < 0)
	return PCL_FAILURE;
      
      /* select events */
      arg = event_select;
      rc = ioctl(hw_descr->fd, PCNT5MUX, &arg);
      if (rc < 0)
	return PCL_FAILURE;
      
      /* select interrupt frequency */
      arg = int_frequency;
      rc = ioctl(hw_descr->fd, PCNT5FREQ, &arg);
      if (rc < 0)
	return PCL_FAILURE;
      
      /* zero and restart selected counters */
      pm_ev5.pmctrs_ev5_cpu = hw_descr->cpu_num;
      pm_ev5.pmctrs_ev5_select = hw_descr->counter_select;
      rc = ioctl(hw_descr->fd, PCNT5RESTART, &pm_ev5.pmctrs_ev5_long);
      if (rc < 0)
	return PCL_FAILURE;
    }

#if defined(OS_HAS_SUPPORT_FOR_21264)
  else
    {
      /* 21264 */
      /* initialize selectors */
      event_select = hw_descr->counter_select = 0;
      
      /* extract selected events */
      if(IS_USED(used_mask, 0))
	{
	  event_select = commands[0];
	  hw_descr->counter_select = PF6_SEL_COUNTER_0;
	}
      if(IS_USED(used_mask, 1))
	{
	  event_select |= commands[1];
	  hw_descr->counter_select |= PF6_SEL_COUNTER_1;
	}
      
      /* select items (= counters) */
      arg = PFM_COUNTERS;
      rc = ioctl(hw_descr->fd, PCNTSETITEMS, &arg);
      if (rc < 0)
	return PCL_FAILURE;
      
      /* clear driver's counters */
      rc = ioctl(hw_descr->fd, PCNTCLEARCNT);
      if (rc < 0)
	return PCL_FAILURE;
      
#if 0   /* doesn't work as I think it should do */
      /* for 21264-based systems, pfm doesn't support the distinction
	 between user and system mode. What we do here is to count only
	 those processes which have performance counting bit enabled. */
      rc = ioctl(hw_descr->fd, PCNTLOGSELECT);
      if (rc < 0)
	return PCL_FAILURE;
#endif

      /* select events */
      arg = event_select;
      rc = ioctl(hw_descr->fd, PCNT6MUX, &arg);
      if (rc < 0)
	return PCL_FAILURE;
      
      /* zero and restart selected counters */
      pm_ev6.pmctrs_ev6_cpu = hw_descr->cpu_num;
      pm_ev6.pmctrs_ev6_select = hw_descr->counter_select;
      rc = ioctl(hw_descr->fd, PCNT6RESTART, &pm_ev6.pmctrs_ev6_long);
      if (rc < 0)
	return PCL_FAILURE;
    }
#endif


  /* get actual processor cycle counter value (to measure elapsed cycles) */
  hw_descr->cycle_counter_value = read_cycle_counter();
  
  /* everything ok */
  return PCL_SUCCESS;
}


/*                                                                            */
/*============================================================================*/
/* Read hardware counters                                                     */

int
PCL_driver_read(
		PCL_DESCR_TYPE d, /* I: descriptor */
		int max_counter,
		PCL_BIT_MASK_TYPE counter_used_mask,
		PCL_CNT_TYPE *counter_values
		)

{
  int rc;
  long cycles, read_cycles;
  struct pfcntrs_ev5 *pfcntrs_ev5;
  union pmctrs_ev5 pmctrs_ev5;
  PCL_BIT_MASK_TYPE used_mask;
#if defined(OS_HAS_SUPPORT_FOR_21264)
  struct pfcntrs_ev6 *pfcntrs_ev6;
  union pmctrs_ev6 pmctrs_ev6;
#endif
  PCL_state_descr *descr;
  HW_state_descr *hw_descr;


  /* get state descriptor */
  descr = d;
  hw_descr = &descr->hw_state_descr;

  /* set up used mask */
  used_mask = counter_used_mask & PCL_mask[max_counter-1];

  /* get actual cycle count from processor cycle counter */
  read_cycles = read_cycle_counter();
  if(read_cycles < hw_descr->cycle_counter_value)
    /* overflow condition */
    cycles = read_cycles + ((1l << 32) - hw_descr->cycle_counter_value);
  else
    cycles = read_cycles - hw_descr->cycle_counter_value;


  /* read out performance counters through driver interface */
  if(descr->event_table == event_table_21164)
    {
      /* 21164 */
      pfcntrs_ev5 = &hw_descr->cntrs_ev5[hw_descr->cpu_num];
      
      /* get cpu counter values */
      pmctrs_ev5.pmctrs_ev5_long = 0;
      rc = ioctl(hw_descr->fd, PCNT5GETCNTRS, &pmctrs_ev5.pmctrs_ev5_long);
      if (rc < 0)
	return PCL_FAILURE;
      
      /* get driver's interrupt values */
      rc = ioctl(hw_descr->fd, PCNT5GETCNT, &pfcntrs_ev5);
      if (rc < 0)
	return PCL_FAILURE;
      
      /* extract counter values
	 (interrupts * timer_frequency + counter_values)
      */
      if(IS_USED(used_mask, 0))
	counter_values[0] = (pfcntrs_ev5->pf_cntr0 << 16)
	  + pmctrs_ev5.pmctrs_ev5_cntr0;
      if(IS_USED(used_mask, 1))
	counter_values[1] = (pfcntrs_ev5->pf_cntr1 << 16)
	  + pmctrs_ev5.pmctrs_ev5_cntr1;
      if(IS_USED(used_mask, 2))
	counter_values[2] = (pfcntrs_ev5->pf_cntr2 << 14)
	  + pmctrs_ev5.pmctrs_ev5_cntr2;
      
      /* special case: elapsed cycles */
      if(IS_USED(used_mask, 3))
	counter_values[3] = cycles;
    }

#if defined(OS_HAS_SUPPORT_FOR_21264)
  else
    {
      /* 21264 */
      /* get cpu counter values */
      pmctrs_ev6.pmctrs_ev6_long = 0;
      rc = ioctl(hw_descr->fd, PCNT6GETCNTRS, &pmctrs_ev6.pmctrs_ev6_long);
      if (rc < 0)
	return PCL_FAILURE;
      
      /* get driver's interrupt values */
      pfcntrs_ev6 = &hw_descr->cntrs_ev6[hw_descr->cpu_num];
      rc = ioctl(hw_descr->fd, PCNT6GETCNT, &pfcntrs_ev6);
      if (rc < 0)
	return PCL_FAILURE;
      
      /* extract counter values
	 (interrupts * timer_frequency + counter_values) */
      /* timer frequency is fixed to 2^20 (1048576) */
      if(IS_USED(used_mask, 0))
	counter_values[0] = (pfcntrs_ev6->pf_cntr0 << 20)
	  + pmctrs_ev6.pmctrs_ev6_cntr0;
      if(IS_USED(used_mask, 1))
	counter_values[1] = (pfcntrs_ev6->pf_cntr1 << 20)
	  + pmctrs_ev6.pmctrs_ev6_cntr1;
      
      /* special case: elapsed cycles */
      if(IS_USED(used_mask, 2))
	counter_values[2] = cycles;
    }
#endif

  /* everything ok */
  return PCL_SUCCESS;
}


/*                                                                            */
/*============================================================================*/
/*                                                                            */


int
PCL_driver_stop(
		PCL_DESCR_TYPE d /* I: descriptor */
		)

{
  int rc;
  union pmctrs_ev5 pm_ev5;
#if defined(OS_HAS_SUPPORT_FOR_21264)
  union pmctrs_ev6 pm_ev6;
#endif
  PCL_state_descr *descr;
  HW_state_descr *hw_descr;


  /* get state descriptor */
  descr = d;
  hw_descr = &descr->hw_state_descr;

  if(descr->event_table == event_table_21164)
    {
      /* 21164 */
      /* disable counters*/
      pm_ev5.pmctrs_ev5_cpu = hw_descr->cpu_num;
      pm_ev5.pmctrs_ev5_select = hw_descr->counter_select;
      rc = ioctl(hw_descr->fd, PCNT5DISABLE, &pm_ev5.pmctrs_ev5_long);
      if (rc < 0)
	return PCL_FAILURE;
    }

#if defined(OS_HAS_SUPPORT_FOR_21264)
  else
    {
      /* 21264 */
      /* disable counters*/
      pm_ev6.pmctrs_ev6_cpu = hw_descr->cpu_num;
      pm_ev6.pmctrs_ev6_select = hw_descr->counter_select;
      rc = ioctl(hw_descr->fd, PCNT6DISABLE, &pm_ev6.pmctrs_ev6_long);
      if (rc < 0)
	return PCL_FAILURE;
    }
#endif
  
  /* everything ok */
  return PCL_SUCCESS;
}


/*                                                                            */
/*============================================================================*/
/* close                                                                      */

int
PCL_driver_close(
		 PCL_DESCR_TYPE d /* I: descriptor */
		 )

{
  PCL_state_descr *descr;
  HW_state_descr *hw_descr;


  /* get state descriptor */
  descr = d;
  hw_descr = &descr->hw_state_descr;

  /* close device */
  if(close(hw_descr->fd) != 0)
    return PCL_FAILURE;

  /* everything ok */
  return PCL_SUCCESS;
}


/*============================================================================*
 *                             that's all folks                               *
 *============================================================================*/
