/*==============================================================================


   Purpose:    low level routines for Pentiums/Linux


   Author:     Rudolf Berrendorf
               Marc Cremer, Maiko Wessel
               University of Applied Sciences Bonn-Rhein-Sieg
               Computer Science Department
               53754 Sankt Augustin
               rudolf.berrendorf@fh-bonn-rhein-sieg.de


   last mod:   22.11.2002


   DISCLAIMER

                  PCL - The Performance Counter Library


   Copyright (C) 2002
   Rudolf Berrendorf

   Copyright (C) 2000
   Forschungszentrum Juelich GmbH, Federal Republic of Germany.
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are met:

   Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

     - Redistributions of source code must retain the above copyright notice,
       this list of conditions and the following disclaimer.

     - Redistributions in binary form must reproduce the above copyright
       notice, this list of conditions and the following disclaimer in the
       documentation and/or other materials provided with the distribution.

     - Any publications that result from the use of this software shall
       reasonably refer to the Research Centre's development.

     - All advertising materials mentioning features or use of this software
       must display the following acknowledgement:

           This product includes software developed by Forschungszentrum
           Juelich GmbH, Federal Republic of Germany.

     - Forschungszentrum Juelich GmbH is not obligated to provide the user with
       any support, consulting, training or assistance of any kind with regard
       to the use, operation and performance of this software or to provide
       the user with any updates, revisions or new versions.


   THIS SOFTWARE IS PROVIDED BY FORSCHUNGSZENTRUM JUELICH GMBH "AS IS" AND ANY
   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
   DISCLAIMED. IN NO EVENT SHALL FORSCHUNGSZENTRUM JUELICH GMBH BE LIABLE FOR
   ANY SPECIAL, DIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
   RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
   CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
   CONNECTION WITH THE ACCESS, USE OR PERFORMANCE OF THIS SOFTWARE.

==============================================================================*/

#include <string.h>
#include <stdio.h>

#include "../../defines.h"
#include "../../pcl.h"
#include "hw_driver_defs.h"
#include "../../pcl_hw_driver.h"
#include "../../hw_driver_global_defs.h"
#include "../../state_descr.h"
#include "../../determine_mhz.h"


/*                                                                            */
/*----------------------------------------------------------------------------*/
/* macros                                                                     */

/* vendor identifier as delivered by CPUID (first 4 bytes only) */
/* Intel: GenuineINTEL */
#define INTEL_ID (('G' << 0) | ('e' << 8) | ('n' << 16) | ('u' << 24))
/* AMD: AuthenticAMD */
#define AMD_ID   (('A' << 0) | ('u' << 8) | ('t' << 16) | ('h' << 24))


/*                                                                            */
/*----------------------------------------------------------------------------*/
/* local variables                                                            */

/* invalid data entry if event is not available */
#define PCL_INVALID  0,{{{-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},\
                         {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}


/*----------------------------------------------------------------------------*/

/* Intel Pentium */
static const PCL_event_descr_t event_table_pentium[PCL_MAX_EVENT] =
{
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_MISS */
  {PCL_AVAIL,NOT_COMP,{{{0x00,0x00,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD,OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_L1DCACHE_READ */
  {PCL_AVAIL,NOT_COMP,{{{0x01,0x01,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD,OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_L1DCACHE_WRITE */
  {PCL_AVAIL,COMPOSED,{{{0x00,0x01,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD,OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_L1DCACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1DCACHE_HIT */
  {PCL_AVAIL,NOT_COMP,{{{0x29,0x29,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD,OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_L1DCACHE_MISS */
  {PCL_AVAIL,NOT_COMP,{{{0x0C,0x0C,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD,OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_L1ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1ICACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1ICACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1ICACHE_HIT */
  {PCL_AVAIL,NOT_COMP,{{{0x0E,0x0E,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD,OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_L1ICACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2CACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2CACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2CACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2CACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2CACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2DCACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2DCACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2DCACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2DCACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2DCACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_TLB_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_TLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_ITLB_HIT */
  {PCL_AVAIL,NOT_COMP,{{{0x0D,0x0D,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD,OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_ITLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_DTLB_HIT */
  {PCL_AVAIL,NOT_COMP,{{{0x02,0x02,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD,OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_DTLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_CYCLES */
  /* special case with dummy entry */
  {PCL_AVAIL,NOT_COMP,{{{-1,-1,0x00,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {0,0,OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}},  /* PCL_ELAPSED_CYCLES */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_INTEGER_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{0x22,0x22,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD,OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_FP_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_LOAD_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STORE_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{0x28,0x28,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD,OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_LOADSTORE_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{0x16,0x16,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD,OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_JUMP_SUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_JUMP_UNSUCCESS */
  {PCL_AVAIL,NOT_COMP,{{{0x12,0x12,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD,OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_JUMP */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_ATOMIC_SUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_ATOMIC_UNSUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_ATOMIC */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STALL_INTEGER */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STALL_FP */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STALL_JUMP */
  {PCL_AVAIL,NOT_COMP,{{{0x1A,0x1A,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD,OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_STALL_LOAD */
  {PCL_AVAIL,NOT_COMP,{{{0x19,0x19,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD,OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_STALL_STORE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STALL */
  {PCL_UNAVAIL, PCL_INVALID},          /* PCL_MFLOPS */
  {PCL_UNAVAIL, PCL_INVALID},          /* PCL_IPC */
  {PCL_AVAIL,COMPOSED,{{{0x29,0x28,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_FP_2|OP_FP_STORE,OP_FP_2|OP_FP_DIV,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_L1DCACHE_MISSRATE */
  {PCL_UNAVAIL, PCL_INVALID},        /* PCL_L2DCACHE_MISSRATE */
  {PCL_AVAIL,COMPOSED,{{{0x28,0x22,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_FP_2|OP_FP_STORE,OP_FP_2|OP_FP_DIV,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}} /* PCL_MEM_FP_RATIO */
};

/*----------------------------------------------------------------------------*/


/* Intel Pentium MMX */
static const PCL_event_descr_t event_table_pentium_mmx[PCL_MAX_EVENT] =
{
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_MISS */
  {PCL_AVAIL,NOT_COMP,{{{0x00,0x00,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD,OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_L1DCACHE_READ */
  {PCL_AVAIL,NOT_COMP,{{{0x01,0x01,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD,OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_L1DCACHE_WRITE */
  {PCL_AVAIL,COMPOSED,{{{0x00,0x01,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD,OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_L1DCACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1DCACHE_HIT */
  {PCL_AVAIL,NOT_COMP,{{{0x29,0x29,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD,OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_L1DCACHE_MISS */
  {PCL_AVAIL,NOT_COMP,{{{0x0C,0x0C,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD,OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_L1ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1ICACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1ICACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1ICACHE_HIT */
  {PCL_AVAIL,NOT_COMP,{{{0x0E,0x0E,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD,OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_L1ICACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2CACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2CACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2CACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2CACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2CACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2DCACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2DCACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2DCACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2DCACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2DCACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_TLB_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_TLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_ITLB_HIT */
  {PCL_AVAIL,NOT_COMP,{{{0x0D,0x0D,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD,OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_ITLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_DTLB_HIT */
  {PCL_AVAIL,NOT_COMP,{{{0x02,0x02,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD,OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_DTLB_MISS */
  {PCL_AVAIL,NOT_COMP,{{{0x30,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_CYCLES */
  /* special case with dummy entry */
  {PCL_AVAIL,NOT_COMP,{{{-1,-1,0x00,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {0,0,OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}},  /* PCL_ELAPSED_CYCLES */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_INTEGER_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{0x22,0x22,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD,OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_FP_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_LOAD_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STORE_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{0x28,0x28,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD,OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_LOADSTORE_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{0x16,0x16,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD,OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{-1,0x31,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {0,OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_JUMP_SUCCESS */
  {PCL_AVAIL,COMPOSED,{{{0x12,0x31,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD,OP_SUB,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_JUMP_UNSUCCESS */
  {PCL_AVAIL,NOT_COMP,{{{0x12,0x12,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD,OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_JUMP */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_ATOMIC_SUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_ATOMIC_UNSUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_ATOMIC */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STALL_INTEGER */
  {PCL_AVAIL,NOT_COMP,{{{0x32,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}},  /* PCL_STALL_FP */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STALL_JUMP */
  {PCL_AVAIL,NOT_COMP,{{{0x1A,0x1A,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD,OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_STALL_LOAD */
  {PCL_AVAIL,NOT_COMP,{{{0x19,0x19,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD,OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_STALL_STORE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STALL */
  {PCL_AVAIL,COMPOSED,{{{0x30,0x22,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_FP_1|OP_FP_2|OP_FP_REZ_MHZ,OP_FP_2|OP_FP_MUL,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_MFLOPS */
  {PCL_AVAIL,COMPOSED,{{{0x30,0x16,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_FP_1|OP_FP_2|OP_FP_REZ,OP_FP_2|OP_FP_MUL,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_IPC */
  {PCL_AVAIL,COMPOSED,{{{0x29,0x28,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_FP_2|OP_FP_STORE,OP_FP_2|OP_FP_DIV,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_L1DCACHE_MISSRATE */
  {PCL_UNAVAIL, PCL_INVALID},          /* PCL_L2DCACHE_MISSRATE */
  {PCL_AVAIL,COMPOSED,{{{0x28,0x22,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_FP_2|OP_FP_STORE,OP_FP_2|OP_FP_DIV,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}} /* PCL_MEM_FP_RATIO */
};


/*----------------------------------------------------------------------------*/


/* Intel Pentium II, III */
static const PCL_event_descr_t event_table_pentium2[PCL_MAX_EVENT] =
{
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1DCACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1DCACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1DCACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1DCACHE_HIT */
  {PCL_AVAIL,NOT_COMP,{{{0x45,0x45,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD,OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_L1DCACHE_MISS */
  {PCL_AVAIL,NOT_COMP,{{{0x80,0x80,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD,OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_L1ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1ICACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1ICACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1ICACHE_HIT */
  {PCL_AVAIL,NOT_COMP,{{{0x81,0x81,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD,OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_L1ICACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2CACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2CACHE_WRITE */
  {PCL_AVAIL,NOT_COMP,{{{0x2E,0x2E,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD,OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_L2CACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2CACHE_HIT */
  {PCL_AVAIL,NOT_COMP,{{{0x24,0x24,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD,OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_L2CACHE_MISS */
  {PCL_AVAIL,NOT_COMP,{{{0x29,0x29,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD,OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_L2DCACHE_READ */
  {PCL_AVAIL,NOT_COMP,{{{0x2A,0x2A,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD,OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_L2DCACHE_WRITE */
  {PCL_AVAIL,COMPOSED,{{{0x29,0x2A,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD,OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_L2DCACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2DCACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2DCACHE_MISS */
  {PCL_AVAIL,NOT_COMP,{{{0x28,0x28,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD,OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_L2ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_TLB_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_TLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_ITLB_HIT */
  {PCL_AVAIL,NOT_COMP,{{{0x85,0x85,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD,OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_ITLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_DTLB_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_DTLB_MISS */
  {PCL_AVAIL,NOT_COMP,{{{0x79,0x79,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD,OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_CYCLES */
  /* special case with dummy entry */
  {PCL_AVAIL,NOT_COMP,{{{-1,-1,0x00,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {0,0,OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}},  /* PCL_ELAPSED_CYCLES */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_INTEGER_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{0xC1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}},  /* PCL_FP_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_LOAD_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STORE_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{0x43,0x43,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD,OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_LOADSTORE_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{0xC0,0xC0,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD,OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{0xC9,0xC9,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD,OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_JUMP_SUCCESS */
  {PCL_AVAIL,NOT_COMP,{{{0xC5,0xC5,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD,OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_JUMP_UNSUCCESS */
  {PCL_AVAIL,NOT_COMP,{{{0xC4,0xC4,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD,OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_JUMP */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_ATOMIC_SUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_ATOMIC_UNSUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_ATOMIC */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STALL_INTEGER */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STALL_FP */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STALL_JUMP */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STALL_LOAD */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STALL_STORE */
  {PCL_AVAIL,NOT_COMP,{{{0xA2,0xA2,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD,OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}},  /* PCL_STALL */
  {PCL_AVAIL,COMPOSED,{{{0xC1,0x79,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_FP_2|OP_FP_STORE,OP_FP_2|OP_FP_DIV_MHZ,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_MFLOPS */
  {PCL_AVAIL,COMPOSED,{{{0xC0,0x79,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_FP_2|OP_FP_STORE,OP_FP_2|OP_FP_DIV,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_IPC */
  {PCL_AVAIL,COMPOSED,{{{0x45,0x43,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_FP_2|OP_FP_STORE,OP_FP_2|OP_FP_DIV,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_L1DCACHE_MISSRATE */
  {PCL_UNAVAIL, PCL_INVALID},          /* PCL_L2DCACHE_MISSRATE */
  {PCL_UNAVAIL, PCL_INVALID}           /* PCL_MEM_FP_RATIO */
};


/*----------------------------------------------------------------------------*/
/* Intel Pentium 4 */

/* These enum identifiers are used in the following event table as an index
   to the detailled event description table for P4's in the P4 section below.
   Some events need more than one counter but are not composed in our sense as
   only one counter generates actually a result. As a consequence we mark the
   other used counters in these events with a special entry P4_DUMMY as used.
*/
enum {
  P4_DUMMY,
  /* L1 Miss */ P4_L1M_12, P4_L1M_13, P4_L1M_14, P4_L1M_15, P4_L1M_16, P4_L1M_17,
  /* L1 I Miss */ P4_L1IM_0, P4_L1IM_1, P4_L1IM_2, P4_L1IM_3, P4_L1IM_4,
  /* L2 Read */ P4_L2R_0, P4_L2R_1, P4_L2R_2, P4_L2R_3, 
  /* L2 Hit */ P4_L2H_0, P4_L2H_1, P4_L2H_2, P4_L2H_3,
  /* L2 Miss */ P4_L2M_0, P4_L2M_1, P4_L2M_2, P4_L2M_3,
  /* ITLB Miss */ P4_ITLBM_0, P4_ITLBM_1, P4_ITLBM_2, P4_ITLBM_3,
  /* DTLB Miss */ P4_DTLBM_0, P4_DTLBM_1, P4_DTLBM_2, P4_DTLBM_3,
  /* Cycles */ P4_CYCLES_12, P4_CYCLES_13, P4_CYCLES_14, P4_CYCLES_15, P4_CYCLES_16, P4_CYCLES_17,
  /* Load instr. */ P4_LOAD_12,
  /* Store instr. */ P4_STORE_12,
  /* Load/store instr. */ P4_LOADSTORE_12, P4_LOADSTORE_14,
  /* Instructions */ P4_INSTR_12, P4_INSTR_13, P4_INSTR_14, P4_INSTR_15, P4_INSTR_16, P4_INSTR_17,
  /* FP instructions */ P4_FP_8,
  /* Jump succ. */ P4_JUMPS_12, P4_JUMPS_13, P4_JUMPS_14, P4_JUMPS_15, P4_JUMPS_16, P4_JUMPS_17,
  /* Jump unsucc. */ P4_JUMPU_12, P4_JUMPU_13, P4_JUMPU_14, P4_JUMPU_15, P4_JUMPU_16, P4_JUMPU_17,
};

/* entries are indices! */
static const PCL_event_descr_t event_table_pentium4[PCL_MAX_EVENT] =
{
  {PCL_UNAVAIL, PCL_INVALID},          /* PCL_L1CACHE_READ*/
  {PCL_UNAVAIL, PCL_INVALID},          /* PCL_L1CACHE_WRITE*/
  {PCL_UNAVAIL, PCL_INVALID},          /* PCL_L1CACHE_READWRITE*/
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_HIT*/
  {PCL_UNAVAIL, PCL_INVALID},          /* PCL_L1CACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1DCACHE_READ*/
  {PCL_UNAVAIL, PCL_INVALID},          /* PCL_L1DCACHE_WRITE*/
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1DCACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1DCACHE_HIT */
  {PCL_AVAIL,NOT_COMP,{{{-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,P4_L1M_12,P4_L1M_13,P4_L1M_14,P4_L1M_15,P4_L1M_16,P4_L1M_17,-1},
                        {0,0,0,0,0,0,0,0,0,0,0,0,OP_ADD,OP_ADD,OP_ADD,OP_ADD,OP_ADD,OP_ADD,0}}}}, /* PCL_L1DCACHE_MISS*/
  {PCL_UNAVAIL, PCL_INVALID},          /* PCL_L1ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1ICACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1ICACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},          /* PCL_L1ICACHE_HIT */
  {PCL_AVAIL,NOT_COMP,{{{P4_L1IM_0,P4_L1IM_1,P4_L1IM_2,P4_L1IM_3,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD,OP_ADD,OP_ADD,OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_L1ICACHE_MISS */
  {PCL_AVAIL,NOT_COMP,{{{P4_L2R_0,P4_L2R_1,P4_L2R_2,P4_L2R_3,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD,OP_ADD,OP_ADD,OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_L2CACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2CACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},          /* PCL_L2CACHE_READWRITE */
  {PCL_AVAIL,NOT_COMP,{{{P4_L2H_0,P4_L2H_1,P4_L2H_2,P4_L2H_3,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD,OP_ADD,OP_ADD,OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_L2CACHE_HIT */
  {PCL_AVAIL,NOT_COMP,{{{P4_L2M_0,P4_L2M_1,P4_L2M_2,P4_L2M_3,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD,OP_ADD,OP_ADD,OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_L2CACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2DCACHE_READ*/
  {PCL_UNAVAIL, PCL_INVALID},          /* PCL_L2DCACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},          /* PCL_L2DCACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2DCACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2DCACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_TLB_HIT */
  {PCL_UNAVAIL, PCL_INVALID},          /* PCL_TLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},          /* PCL_ITLB_HIT */
  {PCL_AVAIL,NOT_COMP,{{{P4_ITLBM_0,P4_ITLBM_1,P4_ITLBM_2,P4_ITLBM_3,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD,OP_ADD,OP_ADD,OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_ITLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},          /* PCL_DTLB_HIT */
  {PCL_AVAIL,NOT_COMP,{{{P4_DTLBM_0,P4_DTLBM_1,P4_DTLBM_2,P4_DTLBM_3,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD,OP_ADD,OP_ADD,OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_DTLB_MISS */
  /* special case with dummy entry */
  {PCL_AVAIL,NOT_COMP,{{{-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,P4_CYCLES_12,P4_CYCLES_13,P4_CYCLES_14,P4_CYCLES_15,P4_CYCLES_16,P4_CYCLES_17,-1},
                        {0,0,0,0,0,0,0,0,0,0,0,0,OP_ADD,OP_ADD,OP_ADD,OP_ADD,OP_ADD,OP_ADD,0}}}}, /* PCL_CYCLES */
  /* special case with dummy entry */
  {PCL_AVAIL,NOT_COMP,{{{-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,1},
                        {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,OP_ADD}}}}, /* PCL_ELAPSED_CYCLES */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_INTEGER_INSTR */
  {PCL_AVAIL,COMPOSED,{{{-1,-1,-1,-1,-1,-1,-1,-1,P4_FP_8,-1,-1,-1,P4_DUMMY,-1,-1,-1,-1,-1,-1},
                        {0,0,0,0,0,0,0,0,OP_ADD,0,0,0,OP_NOOP,0,0,0,0,0,0}}}}, /* PCL_FP_INSTR */
  {PCL_AVAIL,COMPOSED,{{{-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,P4_LOAD_12,P4_DUMMY,-1,-1,-1,-1,-1},
                        {0,0,0,0,0,0,0,0,0,0,0,0,OP_ADD,OP_NOOP,0,0,0,0,0}}}}, /* PCL_LOAD_INSTR */
  {PCL_AVAIL,COMPOSED,{{{-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,P4_STORE_12,P4_DUMMY,-1,-1,-1,-1,-1},
                        {0,0,0,0,0,0,0,0,0,0,0,0,OP_ADD,OP_NOOP,0,0,0,0,0}}}}, /* PCL_STORE_INSTR */
  {PCL_AVAIL,COMPOSED,{{{-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,P4_LOADSTORE_12,P4_DUMMY,-1,-1,-1,-1,-1},
                        {0,0,0,0,0,0,0,0,0,0,0,0,OP_ADD,OP_NOOP,0,0,0,0,0}}}}, /* PCL_LOADSTORE_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,P4_INSTR_12,P4_INSTR_13,P4_INSTR_14,P4_INSTR_15,P4_INSTR_16,P4_INSTR_17,-1},
                        {0,0,0,0,0,0,0,0,0,0,0,0,OP_ADD,OP_ADD,OP_ADD,OP_ADD,OP_ADD,OP_ADD,0}}}}, /* PCL_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,P4_JUMPS_12,P4_JUMPS_13,P4_JUMPS_14,P4_JUMPS_15,P4_JUMPS_16,P4_JUMPS_17,-1},
                        {0,0,0,0,0,0,0,0,0,0,0,0,OP_ADD,OP_ADD,OP_ADD,OP_ADD,OP_ADD,OP_ADD,0}}}}, /* PCL_JUMP_SUCCESS */
  {PCL_AVAIL,NOT_COMP,{{{-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,P4_JUMPU_12,P4_JUMPU_13,P4_JUMPU_14,P4_JUMPU_15,P4_JUMPU_16,P4_JUMPU_17,-1},
                        {0,0,0,0,0,0,0,0,0,0,0,0,OP_ADD,OP_ADD,OP_ADD,OP_ADD,OP_ADD,OP_ADD,0}}}}, /* PCL_JUMP_UNSUCCESS */
  {PCL_AVAIL,COMPOSED,{{{-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,P4_JUMPS_12,-1,P4_JUMPU_14,-1,-1,-1,-1},
                        {0,0,0,0,0,0,0,0,0,0,0,0,OP_ADD,0,OP_ADD,0,0,0,0}}}}, /* PCL_JUMP */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_ATOMIC_SUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_ATOMIC_UNSUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_ATOMIC */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STALL_INTEGER */
  {PCL_UNAVAIL, PCL_INVALID},          /* PCL_STALL_FP */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STALL_JUMP */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STALL_LOAD */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STALL_STORE */
  {PCL_UNAVAIL, PCL_INVALID},          /* PCL_STALL */
  {PCL_AVAIL,COMPOSED,{{{-1,-1,-1,-1,-1,-1,-1,-1,P4_FP_8,-1,-1,-1,P4_DUMMY,-1,-1,-1,-1,P4_CYCLES_17,-1},
                        {0,0,0,0,0,0,0,0,OP_FP_2|OP_FP_STORE,0,0,0,OP_NOOP,0,0,0,0,OP_FP_2|OP_FP_DIV_MHZ,0}}}}, /* PCL_MFLOPS */
  {PCL_AVAIL,COMPOSED,{{{-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,P4_INSTR_14,P4_CYCLES_15,-1,-1,-1},
                        {0,0,0,0,0,0,0,0,0,0,0,0,0,0,OP_FP_2|OP_FP_STORE,OP_FP_2|OP_FP_DIV,0,0,0}}}}, /* PCL_IPC */
  {PCL_UNAVAIL, PCL_INVALID},          /* PCL_L1DCACHE_MISSRATE */
  {PCL_AVAIL,COMPOSED,{{{P4_L2M_0,-1,P4_L2R_2,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_FP_2|OP_FP_STORE,0,OP_FP_2|OP_FP_DIV,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_L2DCACHE_MISSRATE */
  {PCL_AVAIL,COMPOSED,{{{-1,-1,-1,-1,-1,-1,-1,-1,P4_FP_8,-1,-1,-1,P4_DUMMY,-1,P4_LOADSTORE_14,P4_DUMMY,-1,-1,-1},
                        {0,0,0,0,0,0,0,0,OP_FP_2|OP_FP_REZ,0,0,0,OP_NOOP,0,OP_FP_2|OP_FP_MUL,OP_NOOP,0,0,0}}}} /* PCL_MEM_FP_RATIO */
};


/*----------------------------------------------------------------------------*/


/* AMD Athlon, Duron */
static const PCL_event_descr_t event_table_athlon[PCL_MAX_EVENT] =
{
  {PCL_UNAVAIL, PCL_INVALID},          /* PCL_L1CACHE_READ*/
  {PCL_UNAVAIL, PCL_INVALID},          /* PCL_L1CACHE_WRITE*/
  {PCL_UNAVAIL, PCL_INVALID},          /* PCL_L1CACHE_READWRITE*/
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_HIT*/
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_MISS*/
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1DCACHE_READ*/
  {PCL_UNAVAIL, PCL_INVALID},          /* PCL_L1DCACHE_WRITE*/
  {PCL_AVAIL,NOT_COMP,{{{0x40,0x40,0x40,0x40,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD,OP_ADD,OP_ADD,OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_L1DCACHE_READWRITE */
  {PCL_AVAIL,COMPOSED,{{{0x40,0x41,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD,OP_SUB,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_L1DCACHE_HIT */
  {PCL_AVAIL,NOT_COMP,{{{0x41,0x41,0x41,0x41,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD,OP_ADD,OP_ADD,OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_L1DCACHE_MISS */
  {PCL_AVAIL,NOT_COMP,{{{0x80,0x80,0x80,0x80,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD,OP_ADD,OP_ADD,OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_L1ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1ICACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1ICACHE_READWRITE */
  {PCL_AVAIL, COMPOSED,{{{0x80, 0x81,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                         {OP_ADD, OP_SUB,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_L1ICACHE_HIT */
  {PCL_AVAIL,NOT_COMP,{{{0x81,0x81,0x81,0x81,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD,OP_ADD,OP_ADD,OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_L1ICACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},          /* PCL_L2CACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2CACHE_WRITE */
  {PCL_AVAIL,NOT_COMP,{{{0x79,0x79,0x79,0x79,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD,OP_ADD,OP_ADD,OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_L2CACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},          /* PCL_L2CACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},          /* PCL_L2CACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2DCACHE_READ*/
  {PCL_UNAVAIL, PCL_INVALID},          /* PCL_L2DCACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},          /* PCL_L2DCACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2DCACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2DCACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_TLB_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_TLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},          /* PCL_ITLB_HIT */
  {PCL_AVAIL,NOT_COMP,{{{0x85,0x85,0x85,0x85,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD,OP_ADD,OP_ADD,OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_ITLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID}, 	       /* PCL_DTLB_HIT */
  {PCL_AVAIL,NOT_COMP,{{{0x46,0x46,0x46,0x46,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD,OP_ADD,OP_ADD,OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_DTLB_MISS */
  {PCL_AVAIL,NOT_COMP,{{{0x76,0x76,0x76,0x76,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD,OP_ADD,OP_ADD,OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_CYCLES */
  /* special case with dummy entry */
  {PCL_AVAIL,NOT_COMP,{{{-1,-1,-1,-1,0x00,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {0,0,0,0,OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_ELAPSED_CYCLES */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_INTEGER_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},          /* PCL_FP_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_LOAD_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STORE_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},          /* PCL_LOADSTORE_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{0xC0,0xC0,0xC0,0xC0,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD,OP_ADD,OP_ADD,OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{0xC4,0xC4,0xC4,0xC4,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD,OP_ADD,OP_ADD,OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_JUMP_SUCCESS */
  {PCL_AVAIL,NOT_COMP,{{{0xC3,0xC3,0xC3,0xC3,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD,OP_ADD,OP_ADD,OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_JUMP_UNSUCCESS */
  {PCL_AVAIL,NOT_COMP,{{{0xC2,0xC2,0xC2,0xC2,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD,OP_ADD,OP_ADD,OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_JUMP */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_ATOMIC_SUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_ATOMIC_UNSUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_ATOMIC */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STALL_INTEGER */
  {PCL_AVAIL,NOT_COMP,{{{0xd7,0xd7,0xd7,0xd7,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_ADD, OP_ADD, OP_ADD, OP_ADD,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_STALL_FP */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STALL_JUMP */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STALL_LOAD */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STALL_STORE */
  {PCL_UNAVAIL, PCL_INVALID},          /* PCL_STALL */
  {PCL_UNAVAIL, PCL_INVALID},          /* PCL_MFLOPS */
  {PCL_AVAIL,COMPOSED,{{{0xC0,0x76,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_FP_2|OP_FP_STORE,OP_FP_2|OP_FP_DIV,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_IPC */
  {PCL_AVAIL,COMPOSED,{{{0x41,0x40,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                        {OP_FP_2|OP_FP_STORE,OP_FP_2|OP_FP_DIV,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}}}, /* PCL_L1DCACHE_MISSRATE */
  {PCL_UNAVAIL, PCL_INVALID},          /* PCL_L2DCACHE_MISSRATE */
  {PCL_UNAVAIL, PCL_INVALID}           /* PCL_MEM_FP_RATIO */
};


/*                                                                            */
/*============================================================================*/
/* low-level stuff                                                            */


/* fast assembler code to access the cycle counter (TSC) */
/* parameter "val" is modified directly */
#define rdtsc(val) __asm__ __volatile__("rdtsc" : "=A" (val))


/*============================================================================*/
/* A Pentium 4 is rather (no, extremly!) complicated for performance counting
   compared with other Pentiums or compatibles. Rather than storing in the
   event table a command word, we store an index into a table with detailled
   information that understands our low level P4 driver.
*/

#define P4_MAX_REGS 2      /* max. number of CCCR/ESCR pairs used for an event */
#define FAST_RDPMC (1 << 31)  /* we want fast rdpmc */
#define NBOGUS     1
#define DUMMY_COUNTER (-1) /* to mark a counter as dummy (handled elsewhere) */

/* IA32_PEBS_ENABLE structure */
typedef union
{
  unsigned int word;                    /* everything in one word */
  struct
  {
    unsigned int l1:1;                  /* 0: L1 miss */
    unsigned int l2:1;                  /* 1: L2 miss */
    unsigned int dtlb:1;                /* 2: DTLB miss */
    unsigned int reserved1:6;           /* unknown */
    unsigned int b9:1;                  /* 9: MOB load */
    unsigned int b10:1;                 /* 10: split load/store */
    unsigned int reserved2:13;          /* unknown */
    unsigned int uop_tag_flag:1;        /* 24: UOP tag flag */
    unsigned int b25:1;                 /* 25: ??? */
    unsigned int reserved3:6;           /* unknown */
  } pebs;
} pentium4_pebs_t;

/*
   Performance counters on a P4 are controlled by CCRS's and ESCR's.
   The following fields of an CCCR are handled by the driver itself:
   - all reserved fields
 */
typedef union
{
  unsigned int word;              /* everything in one word */
  struct {
    unsigned int reserved1:12;    /* reserved */
    unsigned int enable:1;    	  /* enable counter */
    unsigned int escr_sel:3;      /* escr select */
    unsigned int reserved2:2;     /* reserved must set to 0x03 (hyperthreading) */
    unsigned int compare:1;       /* compare flag */
    unsigned int complement:1;    /* complement flag */
    unsigned int threshold:4;     /* threshold field */
    unsigned int edge:1;     	  /* edge detect */
    unsigned int force_ovf:1;     /* forces overflow */
    unsigned int ovf_pmi:1;       /* generates pmi */
    unsigned int reserved3:3;     /* reserved */
    unsigned int cascade:1;       /* enables cascading */
    unsigned int ovf:1;           /* indicate overflow */
  } cccr;
} pentium4_cccr_t;


typedef int pentium4_pebs_matrix_vert_t;


/* The following fields of an ESCR are handled by the driver itself:
   - all reserved fields
   - usr, os
 */
typedef union
{
  unsigned int word;              /* everything in one word */
  struct
  {
    unsigned int reserved1:2;	  /* reserved */
    unsigned int usr:1;     	  /* count in user mode (cpl=1,2,3) */
    unsigned int os:1;            /* count in os mode (cpl=0) */
    unsigned int tag_enable:1;    /* tag enable */
    unsigned int tag_value:4;     /* tag value */
    unsigned int event_mask:16;   /* event mask (first bit = nbogus bit) */
    unsigned int event_sel:6;     /* event class select */
    unsigned int reserved2:1;     /* reserved */
  } escr;
} pentium4_escr_t;


/* a single CCCR/ESCR descriptor */
typedef struct
{
  int counter_number;             /* counter register number */
  pentium4_cccr_t cccr;           /* CCCR */
  pentium4_escr_t escr;           /* ESCR */
  pentium4_pebs_t pebs;           /* for replay tagging */
  pentium4_pebs_matrix_vert_t pebs_matrix_vert; /* for replay tagging */
} pentium4_descr_t;


/* A P4 event description is at most P4_MAX_REGS single descriptors */
typedef struct
{
  pentium4_descr_t descr[P4_MAX_REGS];
} pentium4_event_descr_t;


/* logical names for ESCR's */
#define MSR_BPU_ESCR0  0
#define MSR_BPU_ESCR1  0
#define MSR_FIRM_ESCR0 1
#define MSR_FIRM_ESCR1 1
#define MSR_TBPU_ESCR0 2
#define MSR_TBPU_ESCR1 2
#define MSR_RAT_ESCR0  2
#define MSR_RAT_ESCR1  2
#define MSR_CRU_ESCR0  4
#define MSR_CRU_ESCR1  4
#define MSR_CRU_ESCR2  5
#define MSR_CRU_ESCR3  5
#define MSR_FSB_ESCR0  6
#define MSR_CR_ESCR0   7
#define MSR_CR_ESCR1   7

/* and here comes our detailled list describing events directly measurable on
   a specific ESCR/CCCR combination */
static const pentium4_event_descr_t pentium4_event_descr[] =
  {
    /*-------------------------------------------------------------------------*/
    /* dummy entry (this is used to code reserved counters) */

    [P4_DUMMY]=
    { 
      .descr[0]=
      { DUMMY_COUNTER,                                    /* counter register */
      },
      .descr[1]= { -1 }
    },

    /*-------------------------------------------------------------------------*/
    /* L1 Data Cache Miss (replay) */

    [P4_L1M_12]=
    { 
      .descr[0]=
      { 12,                                              /* counter register */
	{.cccr.enable=1, .cccr.escr_sel=MSR_CRU_ESCR2},  /* CCCR */
	{.escr.event_mask=NBOGUS, .escr.event_sel=9},    /* ESCR */
	{.pebs.l1=1, .pebs.uop_tag_flag=1},              /* PEBS */
	1                                                /* PEBS matrix vert */
      },
      .descr[1]= { -1 }
    },

    [P4_L1M_13]=
    { 
      .descr[0]=
      { 13,                                              /* counter register */
	{.cccr.enable=1, .cccr.escr_sel=MSR_CRU_ESCR2},  /* CCCR */
	{.escr.event_mask=NBOGUS, .escr.event_sel=9},    /* ESCR */
	{.pebs.l1=1, .pebs.uop_tag_flag=1},              /* PEBS */
	1                                                /* PEBS matrix vert */
      },
      .descr[1]= { -1 }
    },

    [P4_L1M_14]=
    { 
      .descr[0]=
      { 14,                                              /* counter register */
	{.cccr.enable=1, .cccr.escr_sel=MSR_CRU_ESCR3},  /* CCCR */
	{.escr.event_mask=NBOGUS, .escr.event_sel=9},    /* ESCR */
	{.pebs.l1=1, .pebs.uop_tag_flag=1},              /* PEBS */
	1                                                /* PEBS matrix vert */
      },
      .descr[1]= { -1 }
    },

    [P4_L1M_15]=
    { 
      .descr[0]=
      { 15,                                              /* counter register */
	{.cccr.enable=1, .cccr.escr_sel=MSR_CRU_ESCR3},  /* CCCR */
	{.escr.event_mask=NBOGUS, .escr.event_sel=9},    /* ESCR */
	{.pebs.l1=1, .pebs.uop_tag_flag=1},              /* PEBS */
	1                                                /* PEBS matrix vert */
      },
      .descr[1]= { -1 }
    },

    [P4_L1M_16]=
    { 
      .descr[0]=
      { 16,                                              /* counter register */
	{.cccr.enable=1, .cccr.escr_sel=MSR_CRU_ESCR2},  /* CCCR */
	{.escr.event_mask=NBOGUS, .escr.event_sel=9},    /* ESCR */
	{.pebs.l1=1, .pebs.uop_tag_flag=1},              /* PEBS */
	1                                                /* PEBS matrix vert */
      },
      .descr[1]= { -1 }
    },

    [P4_L1M_17]=
    { 
      .descr[0]=
      { 17,                                              /* counter register */
	{.cccr.enable=1, .cccr.escr_sel=MSR_CRU_ESCR3},  /* CCCR */
	{.escr.event_mask=NBOGUS, .escr.event_sel=9},    /* ESCR */
	{.pebs.l1=1, .pebs.uop_tag_flag=1},              /* PEBS */
	1                                                /* PEBS matrix vert */
      },
      .descr[1]= { -1 }
    },


    /*-------------------------------------------------------------------------*/
    /* L1 Instruction Cache Miss (non-retirement) */

    [P4_L1IM_0]=
    { 
      .descr[0]=
      { 0,                                               /* counter register */
	{.cccr.enable=1, .cccr.escr_sel=MSR_BPU_ESCR0},  /* CCCR */
	{.escr.event_mask=1, .escr.event_sel=3}          /* ESCR */
      },
      .descr[1]= { -1 }
    },

    [P4_L1IM_1]=
    { 
      .descr[0]=
      { 1,                                               /* counter register */
	{.cccr.enable=1, .cccr.escr_sel=MSR_BPU_ESCR0},  /* CCCR */
	{.escr.event_mask=1, .escr.event_sel=3}          /* ESCR */
      },
      .descr[1]= { -1 }
    },

    [P4_L1IM_2]=
    { 
      .descr[0]=
      { 2,                                               /* counter register */
	{.cccr.enable=1, .cccr.escr_sel=MSR_BPU_ESCR1},  /* CCCR */
	{.escr.event_mask=1, .escr.event_sel=3}          /* ESCR */
      },
      .descr[1]= { -1 }
    },

    [P4_L1IM_3]=
    { 
      .descr[0]=
      { 3,                                               /* counter register */
	{.cccr.enable=1, .cccr.escr_sel=MSR_BPU_ESCR1},  /* CCCR */
	{.escr.event_mask=1, .escr.event_sel=3}          /* ESCR */
      },
      .descr[1]= { -1 }
    },


    /*-------------------------------------------------------------------------*/
    /* L2 Cache Read (non-retirement) */

    [P4_L2R_0]=
    { 
      .descr[0]=
      { 0,                                              /* counter register */
	{.cccr.enable=1, .cccr.escr_sel=MSR_CR_ESCR0},  /* CCCR */
	{.escr.event_mask=0x107, .escr.event_sel=0xc}   /* ESCR */
      },
      .descr[1]= { -1 }
    },


    [P4_L2R_1]=
    { 
      .descr[0]=
      { 1,                                              /* counter register */
	{.cccr.enable=1, .cccr.escr_sel=MSR_CR_ESCR0},  /* CCCR */
	{.escr.event_mask=0x107, .escr.event_sel=0xc}   /* ESCR */
      },
      .descr[1]= { -1 }
    },


    [P4_L2R_2]=
    { 
      .descr[0]=
      { 2,                                              /* counter register */
	{.cccr.enable=1, .cccr.escr_sel=MSR_CR_ESCR1},  /* CCCR */
	{.escr.event_mask=0x107, .escr.event_sel=0xc}   /* ESCR */
      },
      .descr[1]= { -1 }
    },


    [P4_L2R_3]=
    { 
      .descr[0]=
      { 3,                                              /* counter register */
	{.cccr.enable=1, .cccr.escr_sel=MSR_CR_ESCR1},  /* CCCR */
	{.escr.event_mask=0x107, .escr.event_sel=0xc}   /* ESCR */
      },
      .descr[1]= { -1 }
    },


    /*-------------------------------------------------------------------------*/
    /* L2 Cache Hit (non-retirement) */

    [P4_L2H_0]=
    { 
      .descr[0]=
      { 0,                                              /* counter register */
	{.cccr.enable=1, .cccr.escr_sel=MSR_CR_ESCR0},  /* CCCR */
	{.escr.event_mask=0x007, .escr.event_sel=0xc}   /* ESCR */
      },
      .descr[1]= { -1 }
    },


    [P4_L2H_1]=
    { 
      .descr[0]=
      { 1,                                              /* counter register */
	{.cccr.enable=1, .cccr.escr_sel=MSR_CR_ESCR0},  /* CCCR */
	{.escr.event_mask=0x007, .escr.event_sel=0xc}   /* ESCR */
      },
      .descr[1]= { -1 }
    },


    [P4_L2H_2]=
    { 
      .descr[0]=
      { 2,                                              /* counter register */
	{.cccr.enable=1, .cccr.escr_sel=MSR_CR_ESCR1},  /* CCCR */
	{.escr.event_mask=0x007, .escr.event_sel=0xc}   /* ESCR */
      },
      .descr[1]= { -1 }
    },


    [P4_L2H_3]=
    { 
      .descr[0]=
      { 3,                                              /* counter register */
	{.cccr.enable=1, .cccr.escr_sel=MSR_CR_ESCR1},  /* CCCR */
	{.escr.event_mask=0x007, .escr.event_sel=0xc}   /* ESCR */
      },
      .descr[1]= { -1 }
    },


    /*-------------------------------------------------------------------------*/
    /* L2 Cache Miss (non-retirement) */

    [P4_L2M_0]=
    { 
      .descr[0]=
      { 0,                                              /* counter register */
	{.cccr.enable=1, .cccr.escr_sel=MSR_CR_ESCR0},  /* CCCR */
	{.escr.event_mask=0x100, .escr.event_sel=0xc}   /* ESCR */
      },
      .descr[1]= { -1 }
    },


    [P4_L2M_1]=
    { 
      .descr[0]=
      { 1,                                              /* counter register */
	{.cccr.enable=1, .cccr.escr_sel=MSR_CR_ESCR0},  /* CCCR */
	{.escr.event_mask=0x100, .escr.event_sel=0xc}   /* ESCR */
      },
      .descr[1]= { -1 }
    },


    [P4_L2M_2]=
    { 
      .descr[0]=
      { 2,                                              /* counter register */
	{.cccr.enable=1, .cccr.escr_sel=MSR_CR_ESCR1},  /* CCCR */
	{.escr.event_mask=0x100, .escr.event_sel=0xc}   /* ESCR */
      },
      .descr[1]= { -1 }
    },


    [P4_L2M_3]=
    { 
      .descr[0]=
      { 3,                                              /* counter register */
	{.cccr.enable=1, .cccr.escr_sel=MSR_CR_ESCR1},  /* CCCR */
	{.escr.event_mask=0x100, .escr.event_sel=0xc}   /* ESCR */
      },
      .descr[1]= { -1 }
    },


    /*-------------------------------------------------------------------------*/
    /* ITLB miss (non-retirement) */

    [P4_ITLBM_0]=
    { 
      .descr[0]=
      { 0,                                               /* counter register */
	{.cccr.enable=1, .cccr.escr_sel=MSR_CRU_ESCR0},  /* CCCR */
	{.escr.event_mask=2, .escr.event_sel=1}          /* ESCR */
      },
      .descr[1]= { -1 }
    },

    [P4_ITLBM_1]=
    { 
      .descr[0]=
      { 1,                                               /* counter register */
	{.cccr.enable=1, .cccr.escr_sel=MSR_CRU_ESCR0},  /* CCCR */
	{.escr.event_mask=2, .escr.event_sel=1}          /* ESCR */
      },
      .descr[1]= { -1 }
    },

    [P4_ITLBM_2]=
    { 
      .descr[0]=
      { 2,                                               /* counter register */
	{.cccr.enable=1, .cccr.escr_sel=MSR_CRU_ESCR1},  /* CCCR */
	{.escr.event_mask=2, .escr.event_sel=1}          /* ESCR */
      },
      .descr[1]= { -1 }
    },

    [P4_ITLBM_3]=
    { 
      .descr[0]=
      { 3,                                               /* counter register */
	{.cccr.enable=1, .cccr.escr_sel=MSR_CRU_ESCR1},  /* CCCR */
	{.escr.event_mask=2, .escr.event_sel=1}          /* ESCR */
      },
      .descr[1]= { -1 }
    },


    /*-------------------------------------------------------------------------*/
    /* DTLB miss (non-retirement) */

    [P4_DTLBM_0]=
    { 
      .descr[0]=
      { 0,                                               /* counter register */
	{.cccr.enable=1, .cccr.escr_sel=MSR_CRU_ESCR0},  /* CCCR */
	{.escr.event_mask=1, .escr.event_sel=1}          /* ESCR */
      },
      .descr[1]= { -1 }
    },

    [P4_DTLBM_1]=
    { 
      .descr[0]=
      { 1,                                               /* counter register */
	{.cccr.enable=1, .cccr.escr_sel=MSR_CRU_ESCR0},  /* CCCR */
	{.escr.event_mask=1, .escr.event_sel=1}          /* ESCR */
      },
      .descr[1]= { -1 }
    },

    [P4_DTLBM_2]=
    { 
      .descr[0]=
      { 2,                                               /* counter register */
	{.cccr.enable=1, .cccr.escr_sel=MSR_CRU_ESCR1},  /* CCCR */
	{.escr.event_mask=1, .escr.event_sel=1}          /* ESCR */
      },
      .descr[1]= { -1 }
    },

    [P4_DTLBM_3]=
    { 
      .descr[0]=
      { 3,                                               /* counter register */
	{.cccr.enable=1, .cccr.escr_sel=MSR_CRU_ESCR1},  /* CCCR */
	{.escr.event_mask=1, .escr.event_sel=1}          /* ESCR */
      },
      .descr[1]= { -1 }
    },


    /*-------------------------------------------------------------------------*/
    /* cycles (non-retirement) */
    /* ???this counts non-sleep clock ticks on non-hyperthreaded processors
       See P4 Processor Performance Metrics for a discussion including hyperthreading
    */

    [P4_CYCLES_12]=
    { 
      .descr[0]=
      { 12,                                              /* counter register */
	{.cccr.enable=1, .cccr.escr_sel=MSR_CRU_ESCR2, .cccr.compare=1, .cccr.complement=1, .cccr.threshold=15},   /* CCCR */
	{.escr.event_mask=0, .escr.event_sel=6}          /* ESCR */
      },
      .descr[1]= { -1 }
    },

    [P4_CYCLES_13]=
    { 
      .descr[0]=
      { 13,                                               /* counter register */
	{.cccr.enable=1, .cccr.escr_sel=MSR_CRU_ESCR2, .cccr.compare=1, .cccr.complement=1, .cccr.threshold=15},  /* CCCR */
	{.escr.event_mask=0, .escr.event_sel=6}           /* ESCR */
      },
      .descr[1]= { -1 }
    },

    [P4_CYCLES_14]=
    { 
      .descr[0]=
      { 14,                                               /* counter register */
	{.cccr.enable=1, .cccr.escr_sel=MSR_CRU_ESCR3, .cccr.compare=1, .cccr.complement=1, .cccr.threshold=15},  /* CCCR */
	{.escr.event_mask=0, .escr.event_sel=6}           /* ESCR */
      },
      .descr[1]= { -1 }
    },

    [P4_CYCLES_15]=
    { 
      .descr[0]=
      { 15,                                               /* counter register */
	{.cccr.enable=1, .cccr.escr_sel=MSR_CRU_ESCR3, .cccr.compare=1, .cccr.complement=1, .cccr.threshold=15},  /* CCCR */
	{.escr.event_mask=0, .escr.event_sel=6}           /* ESCR */
      },
      .descr[1]= { -1 }
    },

    [P4_CYCLES_16]=
    { 
      .descr[0]=
      { 16,                                               /* counter register */
	{.cccr.enable=1, .cccr.escr_sel=MSR_CRU_ESCR2, .cccr.compare=1, .cccr.complement=1, .cccr.threshold=15},  /* CCCR */
	{.escr.event_mask=0, .escr.event_sel=6}           /* ESCR */
      },
      .descr[1]= { -1 }
    },

    [P4_CYCLES_17]=
    { 
      .descr[0]=
      { 17,                                               /* counter register */
	{.cccr.enable=1, .cccr.escr_sel=MSR_CRU_ESCR3, .cccr.compare=1, .cccr.complement=1, .cccr.threshold=15},  /* CCCR */
	{.escr.event_mask=0, .escr.event_sel=6}           /* ESCR */
      },
      .descr[1]= { -1 }
    },


    /*-------------------------------------------------------------------------*/
    /* floating point operations (execution event) */

    [P4_FP_8]=
    { 
      .descr[0]=
      { 12,                                               /* counter register */
	{.cccr.enable=1, .cccr.escr_sel=MSR_CRU_ESCR2},   /* CCCR */
	{.escr.event_mask=1, .escr.event_sel=12}          /* ESCR */
      },
      .descr[1]=
      { 8,                                                /* counter register */
	{.cccr.enable=1, .cccr.escr_sel=MSR_FIRM_ESCR0},  /* CCCR */
	{.escr.event_mask=0x8001, .escr.event_sel=4, .escr.tag_enable=1, .escr.tag_value=1} /* ESCR */
      }
    },


    /*-------------------------------------------------------------------------*/
    /* load instructions (front-end event) */

    [P4_LOAD_12]=
    { 
      .descr[0]=
      { 12,                                               /* counter register */
	{.cccr.enable=1, .cccr.escr_sel=MSR_CRU_ESCR2},   /* CCCR */
	{.escr.event_mask=NBOGUS, .escr.event_sel=8}      /* ESCR */
      },
      .descr[1]=
      { 13,                                               /* counter register */
	{.cccr.enable=1, .cccr.escr_sel=MSR_RAT_ESCR0},   /* CCCR */
	{.escr.event_mask=2, .escr.event_sel=2}           /* ESCR */
      }
    },


    /*-------------------------------------------------------------------------*/
    /* store instructions (front-end event) */

    [P4_STORE_12]=
    { 
      .descr[0]=
      { 12,                                               /* counter register */
	{.cccr.enable=1, .cccr.escr_sel=MSR_CRU_ESCR2},   /* CCCR */
	{.escr.event_mask=NBOGUS, .escr.event_sel=8}      /* ESCR */
      },
      .descr[1]=
      { 13,                                               /* counter register */
	{.cccr.enable=1, .cccr.escr_sel=MSR_RAT_ESCR0},   /* CCCR */
	{.escr.event_mask=4, .escr.event_sel=2}           /* ESCR */
      }
    },


    /*-------------------------------------------------------------------------*/
    /* loadstore instructions (front-end event) */

    [P4_LOADSTORE_12]=
    { 
      .descr[0]=
      { 12,                                               /* counter register */
	{.cccr.enable=1, .cccr.escr_sel=MSR_CRU_ESCR3},   /* CCCR */
	{.escr.event_mask=NBOGUS, .escr.event_sel=8}      /* ESCR */
      },
      .descr[1]=
      { 13,                                               /* counter register */
	{.cccr.enable=1, .cccr.escr_sel=MSR_RAT_ESCR1},   /* CCCR */
	{.escr.event_mask=6, .escr.event_sel=2}           /* ESCR */
      }
    },


    [P4_LOADSTORE_14]=
    { 
      .descr[0]=
      { 14,                                               /* counter register */
	{.cccr.enable=1, .cccr.escr_sel=MSR_CRU_ESCR3},   /* CCCR */
	{.escr.event_mask=NBOGUS, .escr.event_sel=8}      /* ESCR */
      },
      .descr[1]=
      { 15,                                               /* counter register */
	{.cccr.enable=1, .cccr.escr_sel=MSR_RAT_ESCR1},   /* CCCR */
	{.escr.event_mask=6, .escr.event_sel=2}           /* ESCR */
      }
    },


    /*-------------------------------------------------------------------------*/
    /* instructions (at-retirement) */

    [P4_INSTR_12]=
    { 
      .descr[0]=
      { 12,                                               /* counter register */
	{.cccr.enable=1, .cccr.escr_sel=MSR_CRU_ESCR0},   /* CCCR */
	{.escr.event_mask=3, .escr.event_sel=2}           /* ESCR */
      },
      .descr[1]= { -1 }
    },

    [P4_INSTR_13]=
    { 
      .descr[0]=
      { 13,                                               /* counter register */
	{.cccr.enable=1, .cccr.escr_sel=MSR_CRU_ESCR0},   /* CCCR */
	{.escr.event_mask=3, .escr.event_sel=2}           /* ESCR */
      },
      .descr[1]= { -1 }
    },

   [P4_INSTR_14]=
    { 
      .descr[0]=
      { 14,                                               /* counter register */
	{.cccr.enable=1, .cccr.escr_sel=MSR_CRU_ESCR1},   /* CCCR */
	{.escr.event_mask=3, .escr.event_sel=2}           /* ESCR */
      },
      .descr[1]= { -1 }
    },

   [P4_INSTR_15]=
    { 
      .descr[0]=
      { 15,                                               /* counter register */
	{.cccr.enable=1, .cccr.escr_sel=MSR_CRU_ESCR1},   /* CCCR */
	{.escr.event_mask=3, .escr.event_sel=2}           /* ESCR */
      },
      .descr[1]= { -1 }
    },

  [P4_INSTR_16]=
    { 
      .descr[0]=
      { 16,                                               /* counter register */
	{.cccr.enable=1, .cccr.escr_sel=MSR_CRU_ESCR0},   /* CCCR */
	{.escr.event_mask=3, .escr.event_sel=2}           /* ESCR */
      },
      .descr[1]= { -1 }
    },

  [P4_INSTR_17]=
    { 
      .descr[0]=
      { 17,                                               /* counter register */
	{.cccr.enable=1, .cccr.escr_sel=MSR_CRU_ESCR1},   /* CCCR */
	{.escr.event_mask=3, .escr.event_sel=2}           /* ESCR */
      },
      .descr[1]= { -1 }
    },


    /*-------------------------------------------------------------------------*/
    /* successful jumps (non-retirement) */

    [P4_JUMPS_12]=
    { 
      .descr[0]=
      { 12,                                               /* counter register */
	{.cccr.enable=1, .cccr.escr_sel=MSR_CRU_ESCR2},   /* CCCR */
	{.escr.event_mask=15, .escr.event_sel=6}          /* ESCR */
      },
      .descr[1]= { -1 }
    },

    [P4_JUMPS_13]=
    { 
      .descr[0]=
      { 13,                                               /* counter register */
	{.cccr.enable=1, .cccr.escr_sel=MSR_CRU_ESCR2},   /* CCCR */
	{.escr.event_mask=15, .escr.event_sel=6}          /* ESCR */
      },
      .descr[1]= { -1 }
    },

    [P4_JUMPS_14]=
    { 
      .descr[0]=
      { 14,                                               /* counter register */
	{.cccr.enable=1, .cccr.escr_sel=MSR_CRU_ESCR3},   /* CCCR */
	{.escr.event_mask=15, .escr.event_sel=6}          /* ESCR */
      },
      .descr[1]= { -1 }
    },

    [P4_JUMPS_15]=
    { 
      .descr[0]=
      { 15,                                               /* counter register */
	{.cccr.enable=1, .cccr.escr_sel=MSR_CRU_ESCR3},   /* CCCR */
	{.escr.event_mask=15, .escr.event_sel=6}          /* ESCR */
      },
      .descr[1]= { -1 }
    },

    [P4_JUMPS_16]=
    { 
      .descr[0]=
      { 16,                                               /* counter register */
	{.cccr.enable=1, .cccr.escr_sel=MSR_CRU_ESCR2},   /* CCCR */
	{.escr.event_mask=15, .escr.event_sel=6}          /* ESCR */
      },
      .descr[1]= { -1 }
    },

    [P4_JUMPS_17]=
    { 
      .descr[0]=
      { 17,                                               /* counter register */
	{.cccr.enable=1, .cccr.escr_sel=MSR_CRU_ESCR3},   /* CCCR */
	{.escr.event_mask=15, .escr.event_sel=6}          /* ESCR */
      },
      .descr[1]= { -1 }
    },


    /*-------------------------------------------------------------------------*/
    /* unsuccessful jumps (non-retirement) */

    [P4_JUMPU_12]=
    { 
      .descr[0]=
      { 12,                                               /* counter register */
	{.cccr.enable=1, .cccr.escr_sel=MSR_CRU_ESCR0},   /* CCCR */
	{.escr.event_mask=1, .escr.event_sel=3}           /* ESCR */
      },
      .descr[1]= { -1 }
    },

    [P4_JUMPU_13]=
    { 
      .descr[0]=
      { 13,                                               /* counter register */
	{.cccr.enable=1, .cccr.escr_sel=MSR_CRU_ESCR0},   /* CCCR */
	{.escr.event_mask=1, .escr.event_sel=3}           /* ESCR */
      },
      .descr[1]= { -1 }
    },

    [P4_JUMPU_14]=
    { 
      .descr[0]=
      { 14,                                               /* counter register */
	{.cccr.enable=1, .cccr.escr_sel=MSR_CRU_ESCR1},   /* CCCR */
	{.escr.event_mask=1, .escr.event_sel=3}           /* ESCR */
      },
      .descr[1]= { -1 }
    },

    [P4_JUMPU_15]=
    { 
      .descr[0]=
      { 15,                                               /* counter register */
	{.cccr.enable=1, .cccr.escr_sel=MSR_CRU_ESCR1},   /* CCCR */
	{.escr.event_mask=1, .escr.event_sel=3}           /* ESCR */
      },
      .descr[1]= { -1 }
    },

    [P4_JUMPU_16]=
    { 
      .descr[0]=
      { 16,                                               /* counter register */
	{.cccr.enable=1, .cccr.escr_sel=MSR_CRU_ESCR0},   /* CCCR */
	{.escr.event_mask=1, .escr.event_sel=3}           /* ESCR */
      },
      .descr[1]= { -1 }
    },

    [P4_JUMPU_17]=
    { 
      .descr[0]=
      { 17,                                               /* counter register */
	{.cccr.enable=1, .cccr.escr_sel=MSR_CRU_ESCR1},   /* CCCR */
	{.escr.event_mask=1, .escr.event_sel=3}           /* ESCR */
      },
      .descr[1]= { -1 }
    },


  };


#define MSR_P4_MISC_ENABLE	0x1A0
#define MSR_P4_MISC_ENABLE_PERF_AVAIL (1<<7)	/* read-only status bit */
#define MSR_P4_MISC_ENABLE_PEBS_UNAVAIL (1<<12) /* read-only status bit */
#define MSR_P4_PERFCTR0		0x300		/* .. 0x311 */
#define MSR_P4_CCCR0		0x360		/* .. 0x371 */
#define MSR_P4_ESCR0		0x3A0		/* .. 0x3E1, with some gaps */

#define MSR_P4_PEBS_ENABLE	0x3F1
#define P4_PE_REPLAY_TAG_BITS	0x00000607
#define P4_PE_UOP_TAG		0x01000000
#define P4_PE_RESERVED		0xFEFFF9F8	/* only allow ReplayTagging */

#define MSR_P4_PEBS_MATRIX_VERT	0x3F2
#define P4_PMV_REPLAY_TAG_BITS	0x00000003
#define P4_PMV_RESERVED		0xFFFFFFFC

#define P4_CCCR_OVF		0x80000000
#define P4_CCCR_CASCADE		0x40000000
#define P4_CCCR_OVF_PMI_T1	0x08000000
#define P4_CCCR_OVF_PMI_T0	0x04000000
#define P4_CCCR_ACTIVE_THREAD	0x00030000
#define P4_CCCR_ENABLE		0x00001000
#define P4_CCCR_ESCR_SELECT(X)	(((X) >> 13) & 0x7)
#define P4_CCCR_RESERVED	(0x30000FFF|P4_CCCR_OVF|P4_CCCR_OVF_PMI_T1)
#define P4_CCCR_REQUIRED	P4_CCCR_ACTIVE_THREAD

#define P4_ESCR_CPL_T1		0x00000003
#define P4_ESCR_CPL_T0		0x0000000C	/* must be non-zero */
#define P4_ESCR_TAG_ENABLE	0x00000010
#define P4_ESCR_RESERVED	(0x80000000|P4_ESCR_CPL_T1)

#define P4_FAST_RDPMC		0x80000000

static const unsigned char p4_cccr_escr_map[4][8] = {
	/* 0x00 and 0x01 as is, 0x02 and 0x03 are +1 */
	[0x00/4] {	[7] 0xA0,
			[6] 0xA2,
			[2] 0xAA,
			[4] 0xAC,
			[0] 0xB2,
			[1] 0xB4,
			[3] 0xB6,
			[5] 0xC8, },
	/* 0x04 and 0x05 as is, 0x06 and 0x07 are +1 */
	[0x04/4] {	[0] 0xC0,
			[2] 0xC2,
			[1] 0xC4, },
	/* 0x08 and 0x09 as is, 0x0A and 0x0B are +1 */
	[0x08/4] {	[1] 0xA4,
			[0] 0xA6,
			[5] 0xA8,
			[2] 0xAE,
			[3] 0xB0, },
	/* 0x0C, 0x0D, and 0x10 as is,
	   0x0E, 0x0F, and 0x11 are +1 except [3] is not in the domain */
	[0x0C/4] {	[4] 0xB8,
			[5] 0xCC,
			[6] 0xE0,
			[0] 0xBA,
			[2] 0xBC,
			[3] 0xBE,
			[1] 0xCA, },
};

static unsigned int new_id(void)
{
	static unsigned int counter;
	int id;

	id = ++counter;
	return id;
}

static unsigned int p4_escr_addr(unsigned int pmc, unsigned int cccr_val)
{
	unsigned int escr_select, pair, escr_offset;

	escr_select = P4_CCCR_ESCR_SELECT(cccr_val);
	if( pmc > 0x11 )
		return 0;	/* pmc range error */
	if( pmc > 0x0F )
		pmc -= 3;	/* 0 <= pmc <= 0x0F */
	pair = pmc / 2;		/* 0 <= pair <= 7 */
	escr_offset = p4_cccr_escr_map[pair / 2][escr_select];
	if( !escr_offset || (pair == 7 && escr_select == 3) )
		return 0;	/* ESCR SELECT range error */
	return escr_offset + (pair & 1) + 0x300;
}

static int p4_check_control(struct perfctr_cpu_control control)
{
	unsigned int i, nractrs, nrctrs, pmc_mask;

	nractrs = control.nractrs;
	nrctrs = nractrs + control.nrictrs;
	if( nrctrs < nractrs || nrctrs > 18 )
		return -1;

	pmc_mask = 0;
	for(i = 0; i < nrctrs; ++i) {
		unsigned int pmc, cccr_val, escr_val, escr_addr;
		/* check that pmc_map[] is well-defined;
		   pmc_map[i] is what we pass to RDPMC, the PMC itself
		   is extracted by masking off the FAST_RDPMC flag */
		pmc = control.pmc_map[i] & ~P4_FAST_RDPMC;
		if( pmc >= 18 || (pmc_mask & (1<<pmc)) )
			return -1;
		pmc_mask |= (1<<pmc);
		/* check CCCR contents */
		cccr_val = control.evntsel[i];
		if( cccr_val & P4_CCCR_RESERVED )
			return -2;
		if( (cccr_val & P4_CCCR_REQUIRED) != P4_CCCR_REQUIRED )
			return -1;
		if( !(cccr_val & (P4_CCCR_ENABLE | P4_CCCR_CASCADE)) )
			return -1;
		if( cccr_val & P4_CCCR_OVF_PMI_T0 ) {
			if( i < nractrs )
				return -1;
		} else {
			if( i >= nractrs )
				return -1;
		}
		/* check ESCR contents */
		escr_val = control.evntsel_aux[i];
		if( escr_val & P4_ESCR_RESERVED )
			return -2;
		if( !(escr_val & P4_ESCR_CPL_T0) )
			return -1;
		/* compute and cache ESCR address */
		escr_addr = p4_escr_addr(pmc, cccr_val);
		if( !escr_addr )
			return -1;		/* ESCR SELECT range error */
		/* XXX: Two counters could map to the same ESCR. Should we
		   check that they use the same ESCR value? */
		//state->k2.p4_escr_map[i] = escr_addr - MSR_P4_ESCR0;
	}
	/* check ReplayTagging control (PEBS_ENABLE and PEBS_MATRIX_VERT) */
	if( control.p4.pebs_enable ) {
		if( !nrctrs )
			return -2;
		if( control.p4.pebs_enable & P4_PE_RESERVED )
			return -2;
		if( !(control.p4.pebs_enable & P4_PE_UOP_TAG ) )
			return -1;
		if( !(control.p4.pebs_enable & P4_PE_REPLAY_TAG_BITS) )
			return -1;
		if( control.p4.pebs_matrix_vert & P4_PMV_RESERVED )
			return -2;
		if( !(control.p4.pebs_matrix_vert & P4_PMV_REPLAY_TAG_BITS) )
			return -1;
	} else if( control.p4.pebs_matrix_vert )
		return -2;
	//state->k1.id = new_id();
	return 0;
}

/*============================================================================*/
/* some utilify functions for debugging */


/* print out relevant fields of a perfctr cpu control structure */
static void
show_control(struct vperfctr_control control)
{
  int i;
  pentium4_cccr_t cccr;
  pentium4_escr_t escr;
  pentium4_pebs_t pebs;
  pentium4_pebs_matrix_vert_t pmv;


  printf("\ncpu control:\n");
  printf("tsc_on                   : %d\n", control.cpu_control.tsc_on);
  printf("nractrs(#a-mode counters): %d\n", control.cpu_control.nractrs);
  printf("nrictrs(#i-mode counters): %d\n", control.cpu_control.nrictrs);
  for(i=0; i<control.cpu_control.nractrs; i++)
    {
      /* CCCR */
      printf("CCCR %d : ", control.cpu_control.pmc_map[i] & ~FAST_RDPMC);
      cccr.word = control.cpu_control.evntsel[i];
      printf("enable=%x, escr_sel=%x, compare=%x, complement=%x, "
	     "threshold=%x, edge=%x, force_ovf=%x, ovf_pmi=%x, "
	     "cascade=%x, ovf=%x",
	     cccr.cccr.enable, cccr.cccr.escr_sel, cccr.cccr.compare,
	     cccr.cccr.complement, cccr.cccr.threshold, cccr.cccr.edge,
	     cccr.cccr.force_ovf, cccr.cccr.ovf_pmi, cccr.cccr.cascade,
	     cccr.cccr.ovf);
      printf("\n");

      /* ESCR */
      printf("ESCR %d : ", control.cpu_control.pmc_map[i] & ~FAST_RDPMC);
      escr.word = control.cpu_control.evntsel_aux[i];
      printf("usr=%x, os=%x, tag_enable=%x, tag_value=%x, event_mask=%x, "
	     "event_sel=%x\n",
	     escr.escr.usr, escr.escr.os, escr.escr.tag_enable, escr.escr.tag_value,
	     escr.escr.event_mask, escr.escr.event_sel);

      /* PEBS */
      printf("PEBS %d : ", control.cpu_control.pmc_map[i] & ~FAST_RDPMC);
      pebs.word = control.cpu_control.p4.pebs_enable;
      printf("l1=%d, l2=%d, dtlb=%d, reserved1=%d, b9=%d, b10=%d, reserved2=%d, "
	     "uop_tag=%d, b25=%d, reserved3=%d\n",
	     pebs.pebs.l1, pebs.pebs.l2, pebs.pebs.dtlb, pebs.pebs.reserved1,
	     pebs.pebs.b9, pebs.pebs.b10, pebs.pebs.reserved2, pebs.pebs.uop_tag_flag,
	     pebs.pebs.b25, pebs.pebs.reserved3);

      /* PEBS matrix vert */
      pmv = control.cpu_control.p4.pebs_matrix_vert;
      printf("PEBS matrx vert %d : %d\n",
	     control.cpu_control.pmc_map[i] & ~FAST_RDPMC,
	     pmv);
    }
}


/*                                                                            */
/*============================================================================*/
/* get information on processor                                               */

int
PCL_driver_info(
                 PCL_PROCESSOR_INFO *info  /* I: address of info struct */
               )

{
  unsigned int signature, feature_flags, family, model, vendor;
  char *(numbers[])= {"0", "1", "2", "3", "4", "5", "6", "7", "8", "9"};


  /* Read first part of vendor id-string from regebx
     eax must be set to 0
     ebx - first part
     edx - second part
     ecx - third part
  */
  vendor = 0;
  asm("cpuid"
      : "=b"(vendor)	/* vendor id string first part */
      : "a"(0));	/* input eax = 0 */

  /* to distinguish between Intel and AMD CPUs, the first part of the
     vendor id string will be compared with "Genu" for "GenuineINTEL"
     or "Auth" for "AuthenticAMD"
  */
  if (vendor == INTEL_ID)
    info->vendor = "Intel";
  else if (vendor == AMD_ID)
    info->vendor = "AMD";
  else
    info->vendor = "unknown";

  /* get all processor type information with the CPUID instruction
     Calling CPUID with EAX=1 returns in EAX version information (type, family,
     model, stepping ID) and in EDX feature information.
     The information in EAX is:
     bit  0-3 : stepping
     bit  4-7 : model
     bit  8-11: family
                0100 = i486
                0101 = Pentium, Pentium MMX
                0110 = Intel: Pentium Pro, Pentium II, Pentium III
                       AMD:   Athlon, Duron
                1111 = Pentium 4
     bit 12-13: processor type
     The information in EDX is (only relevant information shown here):
     bit 4 : Time Stamp Counter supported
     bit 5 : Model Specific Registers implemented with RDMSR, WRMSR instructions
     bit 23: MMX supported
   */
  asm("cpuid"
      : "=a"(signature), "=d"(feature_flags)
      : "a"(1)
      : "%ebx", "%ecx");

  /* now test what processor vendor and what family we have  */
  family = (signature >> 8) & 0xf;
  model = (signature >> 4) & 0xf;
  
#if 0
{
  union {
    unsigned int word;
    struct {
      unsigned int stepping:4;
      unsigned int model:4;
      unsigned int family:4;
      unsigned int type:4;
    } details;
  } cpuid_eax;
  union {
    unsigned int word;
    struct {
      unsigned int reserved1:4;
      unsigned int tsc:1;
      unsigned int rdmsr:1;
      unsigned int reserved2:17;
      unsigned int mmx:1;
    } details;
  } cpuid_edx;

  cpuid_eax.word = signature;
  cpuid_edx.word = feature_flags;
  printf("\n");
  printf ("model    : %d\n", cpuid_eax.details.model);
  printf ("family   : %d\n", cpuid_eax.details.family);
  printf ("stepping : %d\n", cpuid_eax.details.stepping);
  printf ("type     : %d\n", cpuid_eax.details.type);
  printf ("tsc      : %s\n", (cpuid_edx.details.tsc)?"yes":"no");
  printf ("rdmsr    : %s\n", (cpuid_edx.details.rdmsr)?"yes":"no");
  printf ("mmx      : %s\n", (cpuid_edx.details.mmx)?"yes":"no");
}
#endif

  if (vendor == INTEL_ID)
    {
      switch(family)
	{
	case 5:
	  /* Pentium processor; test now for Pentium/MMX */
	  if(((feature_flags >> 23) & 0x1) == 0)
	    /* Pentium without MMX */
	    info->family = "Pentium";
	  else
	    /* Pentium MMX */
	    info->family = "Pentium-MMX";
	  info->ncounters = 3;
	  break;
	  
	case 6:
	  /* Pentium Pro, Pentium II, Pentium III */
	  if(model == 1)
	    info->family = "PPro";
	  else if(model <= 5)
	    info->family = "Pentium II";
	  else if(model <= 9)
	    info->family = "Pentium III";
	  else
	    info->family = "unknown";
	  info->ncounters = 3;
	  break;

	case 15:
	  /* Pentium 4 */
	  info->family = "Pentium 4";
	  info->ncounters = 19;
	  break;
	  
	default:
	  info->family = "unknown";
	}
    }

  else if(vendor == VENDOR_AMD)
    {
      switch(family)
	{
	case 6:
	  /* Athlon, Duron */
	  info->family = "Athlon";
	  info->ncounters = 5;
	  break;

	default:
	  info->family = "unknown";
	}
    }

  else
    /* something we can't handle yet */
    info->family = "unknown";


  /* assign model number */
  if(model < 10)
    info->model = numbers[model];
  else
    info->model = ">9";

  /* get processor speed */
  info->mhz = PCL_determine_mhz_rate();

  return PCL_SUCCESS;
}


/*                                                                            */
/*============================================================================*/
/* We determine if the processor supports performance counters (special MSR)
   and what events are supported. Therefore we use the CPUID instruction to
   test certain features.
   Sets "PCL_event_table" accordingly to get the right register settings.
   Returns PCL_SUCCESS if processor type is supported.
*/

int
PCL_driver_open(
		PCL_DESCR_TYPE d /* I: descriptor */
		)

{
  int res;
  unsigned int signature, feature_flags, family, regebx;
  PCL_state_descr *descr;
  HW_state_descr *hw_descr;
  PCL_PROCESSOR_INFO info;


  /* get state descriptor */
  descr = d;
  hw_descr = &descr->hw_state_descr;

  /* initialize global state variables */
  descr->allowed_count_mode = PCL_MODE_USER_SYSTEM;

  /* get general information on the processor */
  if((res = PCL_driver_info(&info)) != PCL_SUCCESS)
    return res;

  /* now get extra information */
  /* Read first part of vendor id-string from regebx
     eax must be set to 0
     ebx - first part
     ecx - third part
     edx - second part
  */
  regebx = 0;
  asm("cpuid"
      : "=b"(regebx)	/* vendor id string first part */
      : "a"(0));	/* input eax = 0 */

  /* to distinguish between Intel and AMD CPUs, the first part of the vendor id string
     will be compared with "Genu" for "GenuineINTEL" or "Auth" for "AuthenticAMD" */
  if (regebx == INTEL_ID)
    hw_descr->vendor = VENDOR_INTEL;
  else if (regebx == AMD_ID)
    hw_descr->vendor = VENDOR_AMD;
  else
    {
      hw_descr->vendor = VENDOR_UNKNOWN;
      return PCL_NOT_SUPPORTED;
    }

  /* get all processor type information with the CPUID instruction
     Calling CPUID with EAX=1 returns in EAX version information (type, family,
     model, stepping ID) and in EDX feature information.
     The information in EAX is:
     bit  0-3 : stepping
     bit  4-7 : model
     bit  8-11: family
                0100 = i486
                0101 = Pentium
                0110 = Intel: Pentium Pro or Pentium II or Pentium III
                       AMD:   Athlon or Duron
                1111 = Pentium 4
     bit 12-13: processor type
     The information in EDX is (only relevant information shown here):
     bit 4 : Time Stamp Counter supported
     bit 5 : Model Specific Registers implemented with RDMSR, WRMSR instructions
     bit 23: MMX supported
   */
  asm("cpuid"
      : "=a"(signature), "=d"(feature_flags)
      : "a"(1)
      : "%ebx", "%ecx");

  /* check for existence of time stamp counter (TSC) */
  if((((feature_flags >> 4) & 0x1) != 1)
     /* and check for existence of model specific registers (MSR) */
     || (((feature_flags >> 5) & 0x1) != 1))
    /* something old without support for performance counters */
    return PCL_NOT_SUPPORTED;

  /* now test what processor vendor and what family
     we have to choose the right event table */
  family = (signature >> 8) & 0xf;

  if(hw_descr->vendor == VENDOR_INTEL)
    {
      switch(family)
	{
	  case 5:
	    /* Pentium processor; test now for Pentium/MMX */
	    if(((feature_flags >> 23) & 0x1) == 0)
	      {
		/* Pentium without MMX */
		descr->event_table = event_table_pentium;
	      }
	    else
	      /* Pentium MMX */
	      descr->event_table = event_table_pentium_mmx;
	    break;
      
	case 6:
	  /* Pentium Pro, Pentium II, Pentium III */
	  descr->event_table = event_table_pentium2;
	  break;

	case 15:
	  /* Pentium 4 */
	  descr->event_table = event_table_pentium4;
	  break;

	default:
	  return PCL_NOT_SUPPORTED;
	}
    }

  else if(hw_descr->vendor == VENDOR_AMD)
    {
      switch(family)
	{
	case 6:
	  /* Athlon, Duron */
	  descr->event_table = event_table_athlon;
	  break;

	default:
	  return PCL_NOT_SUPPORTED;
	}
    }

  else
    /* something we can't handle yet */
    return PCL_NOT_SUPPORTED;

   /* open /dev/perfctr for read/write */
  if((hw_descr->self = vperfctr_open()) == NULL)
    return PCL_FAILURE;

  /* gcc has some problems in optimization. Therefore we repeat it here */
  /* initialize global state variables */
  descr->allowed_count_mode = PCL_MODE_USER_SYSTEM;

  /* everything ok */
  return PCL_SUCCESS;
}


/*                                                                            */
/*============================================================================*/
/* Starts hardware counters.                                                  */
/* max_counter (> 0) gives the maximum number of counters                     */

int
PCL_driver_start(
		 PCL_DESCR_TYPE d, /* I: descriptor */
		 int max_counter,
		 PCL_BIT_MASK_TYPE counter_used_mask,
		 PCL_DRIVER_COMMAND_TYPE *commands,
		 unsigned int count_mode
		 )

{
  int sel0, sel1, sel2, sel3, cpl, j, res;
  PCL_BIT_MASK_TYPE used_mask;
  PCL_state_descr *descr;
  HW_state_descr *hw_descr;
  struct vperfctr_control control;


  /* get state descriptor */
  descr = d;
  hw_descr = &descr->hw_state_descr;

  /* set up used mask */
  used_mask = counter_used_mask & PCL_mask[max_counter-1];

  /* check count mode */
  if(count_mode == PCL_MODE_USER)
    /* count only in user mode */
    cpl = 2;
  else if(count_mode == PCL_MODE_SYSTEM)
    /* count only system mode */
    cpl = 1;
  else
    /* count in system or user mode */
    cpl = 3;


  /*--------------------------------------------------------------------------*/
  /* Pentium 4 */

  if (descr->event_table == event_table_pentium4)
    {
      int i, k, index;
      pentium4_cccr_t cccr, cccr_mask;
      pentium4_escr_t escr, escr_mask;
      pentium4_descr_t p4;

      /* initialize cccr mask */
      cccr_mask.word = 0;
      /* enable counting */
      cccr_mask.cccr.enable = 1;
      /* set cccr.reserved2 - must be 11B (for Hyperthreading) */
      cccr_mask.cccr.reserved2 = 0x03;

      /* initialize escr mask */
      escr_mask.word = 0;
      /* set privilege level */
      escr_mask.escr.usr = (cpl & 0x2) >> 1;
      escr_mask.escr.os = cpl & 0x1;

      /* intialize control */
      memset(&control, 0, sizeof(control));      


      /* set events */
      i = -1;
      /* we have 18 real performance registers */
      for(j = 0; j < 18; j++)
	{
	  /* counter j used? */
	  if(IS_USED(used_mask, j))
	    {
	      /* command word is actually an index into a description table */
	      index = commands[j];

	      for(k = 0;
		  (k < P4_MAX_REGS)
		    && (pentium4_event_descr[index].descr[k].counter_number != -1);
		  k++)
		{
		  if(pentium4_event_descr[index].descr[k].counter_number != DUMMY_COUNTER)
		    {
		      p4 = pentium4_event_descr[index].descr[k];

		      /* get cccr and escr's */
		      cccr.word = p4.cccr.word | cccr_mask.word;
		      escr.word = p4.escr.word | escr_mask.word;
		  
		      /* one more event to measure */
		      i++;
		      /* PMC mapping */
		      control.cpu_control.pmc_map[i] = p4.counter_number | FAST_RDPMC;
		      /* cccr control word */
		      control.cpu_control.evntsel[i] = cccr.word;
		      /* escr control word */
		      control.cpu_control.evntsel_aux[i] = escr.word;
		      control.cpu_control.p4.pebs_enable = p4.pebs.word;
		      control.cpu_control.p4.pebs_matrix_vert = p4.pebs_matrix_vert;
		    }
		}
	    }
	}

      /* although we don't need the TSC, perfctr is significantly faster
         if TSC is turned on */
      control.cpu_control.tsc_on = 1;       /* turn TSC on */
      control.cpu_control.nractrs = i+1;    /* # a-mode counters */
      
#if defined(DEBUG)
      show_control(control);
      p4_check_control(control.cpu_control);
#endif

      /* call low level interface */
      if((res = vperfctr_control(hw_descr->self, &control)) < 0)
	return PCL_FAILURE;
    }
  
  /*--------------------------------------------------------------------------*/
  /* Athlon, Duron */
  
  else if (descr->event_table == event_table_athlon)
    {
      union
      {
	struct athlon_cesr
	{
	  unsigned int event_select:8;  /* event select */
	  unsigned int unit_mask:8;     /* further qualifies event (MESI) */
	  unsigned int usr:1;           /* count in user mode (cpl=1,2,3) */
	  unsigned int os:1;            /* count in os mode (cpl=0) */
	  unsigned int e:1;             /* edge detect */
	  unsigned int pc:1;            /* pin control */
	  unsigned int aint:1;          /* local APIC interrupt enable on overflow */
	  unsigned int reserved:1;      /* reserved */
	  unsigned int en:1;            /* enable counters (sel0 only) */
	  unsigned int inv:1;           /* invert counter mask */
	  unsigned int counter_mask:8;  /* if!0, compare with events */
	} cesr;
	unsigned int word;              /* to initialize in one assignment */
      } cesr0, cesr1, cesr2, cesr3;


      /* set all bits to 0 */
      cesr0.word = cesr1.word = cesr2.word = cesr3.word = 0;

      /* additional fields */
      cesr0.cesr.usr = cesr1.cesr.usr = cesr2.cesr.usr = cesr3.cesr.usr
	= (cpl & 0x2) >> 1;
      cesr0.cesr.os = cesr1.cesr.os = cesr2.cesr.os = cesr3.cesr.os = cpl & 0x1;
      cesr0.cesr.en = cesr1.cesr.en = cesr2.cesr.en = cesr3.cesr.en = 1;
      /* all other fields get 0 by the above initialization */

      /* preset event selection to some meaningful value (PCL_L1DCACHE_MISS) */
      cesr0.cesr.event_select = 0x41;
      cesr1.cesr.event_select = 0x41;
      cesr2.cesr.event_select = 0x41;
      cesr3.cesr.event_select = 0x41;
 
      if(IS_USED(used_mask, 0))
	{
	  /* events */
	  cesr0.cesr.event_select = sel0 = commands[0];
	  /* count events for all MESI states (applies only to some events) */
	  if((sel0 == 0x42) || (sel0 == 0x43)
	     || (sel0 == 0x44))
	    cesr0.cesr.unit_mask = 0x1e;
	}

      if(IS_USED(used_mask, 1))
	{
	  /* events */
	  cesr1.cesr.event_select = sel1 = commands[1];
	  /* count events for all MESI states (applies only to some events) */
	  if((sel1 == 0x42) || (sel1 == 0x43)
	     || (sel1 == 0x44))
	    cesr1.cesr.unit_mask = 0x1e;
	}
 
      if(IS_USED(used_mask, 2))
	{
	  /* events */
	  cesr2.cesr.event_select = sel2 = commands[2];
	  /* count events for all MESI states (applies only to some events) */
	  if((sel2 == 0x42) || (sel2 == 0x43)
	     || (sel2 == 0x44))
	    cesr2.cesr.unit_mask = 0x1e;
	}

      if(IS_USED(used_mask, 3))
	{
	  /* events */
	  cesr3.cesr.event_select = sel3 = commands[3];
	  /* count events for all MESI states (applies only to some events) */
	  if((sel3 == 0x42) || (sel3 == 0x43)
	     || (sel3 == 0x44))
	    cesr3.cesr.unit_mask = 0x1e;
	}

      /* initialize control structure */
      memset(&control, 0, sizeof control);
      /* although we don't need the TSC, perfctr is significantly faster
         if TSC is turned on */
      control.cpu_control.tsc_on = 1;                 /* turn TSC on */
      control.cpu_control.nractrs = 4;                /* # a-mode counters */
      control.cpu_control.pmc_map[0] = 0;             /* PMC mapping */
      control.cpu_control.pmc_map[1] = 1;             /* PMC mapping */
      control.cpu_control.pmc_map[2] = 2;             /* PMC mapping */
      control.cpu_control.pmc_map[3] = 3;             /* PMC mapping */
      control.cpu_control.evntsel[0] = cesr0.word;    /* control word */
      control.cpu_control.evntsel[1] = cesr1.word;    /* control word */
      control.cpu_control.evntsel[2] = cesr2.word;    /* control word */
      control.cpu_control.evntsel[3] = cesr3.word;    /* control word */

      if(vperfctr_control(hw_descr->self, &control) < 0)
	return PCL_FAILURE;
    }

  /*--------------------------------------------------------------------------*/
  /* Pentium II, III */

  else if (descr->event_table == event_table_pentium2)
    {
      union
      {
	struct pentium2_cesr
	{
	  unsigned int event_select:8;  /* event select */
	  unsigned int unit_mask:8;     /* further qualifies event (MESI) */
	  unsigned int usr:1;           /* count in user mode (cpl=1,2,3) */
	  unsigned int os:1;            /* count in os mode (cpl=0) */
	  unsigned int e:1;             /* edge detect */
	  unsigned int pc:1;            /* pin control */
	  unsigned int aint:1;          /* local APIC interrupt enable on
					   overflow */
	  unsigned int reserved:1;      /* reserved */
	  unsigned int en:1;            /* enable counters (sel0 only) */
	  unsigned int inv:1;           /* invert counter mask */
	  unsigned int counter_mask:8;  /* if!0, compare with events */
	} cesr;
	unsigned int word;              /* to initialize in one assignment */
      } cesr0, cesr1;

      /* set all bits to 0 */
      cesr0.word = cesr1.word = 0;

      /* additional fields */
      cesr0.cesr.usr = cesr1.cesr.usr = (cpl & 0x2) >> 1;
      cesr0.cesr.os = cesr1.cesr.os = cpl & 0x1;
      cesr0.cesr.en = 1;
      /* all other fields get 0 by the above initialization */

      /* preset event selection to some meaningful value (PCL_L1DCACHE_MISS) */
      cesr0.cesr.event_select = 0x45;
      cesr1.cesr.event_select = 0x45;

      if(IS_USED(used_mask, 0))
	{
	  /* events */
	  cesr0.cesr.event_select = sel0 = commands[0];
	  /* count events for all MESI states (applies only to some events) */
	  if((sel0 == 0x28) || (sel0 == 0x29)
	     || (sel0 == 0x2a) || (sel0 == 0x2e))
	    cesr0.cesr.unit_mask = 0xf;
	}

      if(IS_USED(used_mask, 1))
	{
	  cesr1.cesr.event_select = sel1 = commands[1];
	  if((sel1 == 0x28) || (sel1 == 0x29)
	     || (sel1 == 0x2a) || (sel1 == 0x2e))
	    cesr1.cesr.unit_mask = 0xf;
	}


      /* initialize control structure */
      memset(&control, 0, sizeof control);
      /* although we don't need the TSC, perfctr is significantly faster
         if TSC is turned on */
      control.cpu_control.tsc_on = 1;                 /* turn TSC on */
      control.cpu_control.nractrs = 2;                /* # a-mode counters */
      control.cpu_control.pmc_map[0] = 0;             /* PMC mapping */
      control.cpu_control.pmc_map[1] = 1;             /* PMC mapping */
      control.cpu_control.evntsel[0] = cesr0.word;    /* control word */
      control.cpu_control.evntsel[1] = cesr1.word;    /* control word */
      if(vperfctr_control(hw_descr->self, &control) < 0)
	return PCL_FAILURE;
    }

  /*--------------------------------------------------------------------------*/
  /* Pentium, Pentium MMX */

  else if ((descr->event_table == event_table_pentium)
      || (descr->event_table == event_table_pentium_mmx))
    {
      /* sub-fields in the Control and Event Select Register (CESR)
       *     CC0, CC1: CPL Level to Monitor, possibilities are
       *               000 = Count Nothing
       *               001 = Count Event while CPL = 0,1,2
       *               010 = Count Event while CPL = 3
       *               011 = Count Event while CPL = 0,1,2,3
       *               100 = Count Nothing
       *               101 = Count Clocks while CPL = 0,1,2
       *               110 = Count Clocks while CPL = 3
       *               111 = Count Clocks while CPL = 0,1,2,3
       *               Here we only use 000, 001, 010 and 011.
       */
      union
      {
	struct pentium_cesr
	{
	  unsigned int es0:6;        /* event select counter 0 */
	  unsigned int cc0:3;        /* counter control 0 (see above) */
	  unsigned int pc0:1;        /* pin control bit 0
					0=event increment, 1=event overflow */
	  unsigned int reserved1:6;  /* reserved */
	  unsigned int es1:6;        /* event select counter 1 */
	  unsigned int cc1:3;        /* counter control 1 (see cc0) */
	  unsigned int pc1:1;        /* pin control bit 1 */
	  unsigned int reserved2:6;  /* reserved */
	} cesr;
	unsigned int word;           /* to initialize in one assignment */
      } cesr;
      
      /* Extract selected events */
      sel0 = sel1 = 0;
      if(IS_USED(used_mask, 0))
	sel0 = commands[0];
      if(IS_USED(used_mask, 1))
	sel1 = commands[1];

      /* set CESR */
      cesr.word = 0;
      cesr.cesr.es0 = sel0;
      cesr.cesr.es1 = sel1;
      cesr.cesr.cc0 = cesr.cesr.cc1 = cpl;
      
      /* initialize control structure */
      memset(&control, 0, sizeof control);
      /* although we don't need the TSC, perfctr is significantly faster
         if TSC is turned on */
      control.cpu_control.tsc_on = 1;                 /* turn TSC on */
      control.cpu_control.nractrs = 2;                /* # a-mode counters */
      control.cpu_control.pmc_map[0] = 0;             /* PMC mapping */
      control.cpu_control.pmc_map[1] = 1;             /* PMC mapping */
      control.cpu_control.evntsel[0] = cesr.word;     /* control word */
      if(vperfctr_control(hw_descr->self, &control) < 0)
	return PCL_FAILURE;
    }


  /*--------------------------------------------------------------------------*/
  
  else
    /* internal error: something wrong */
    return PCL_FAILURE;

  /*--------------------------------------------------------------------------*/

  /* get actual clock value (to measure elapsed cycles) */
  rdtsc(hw_descr->clock_value);


  return PCL_SUCCESS;
}


/*                                                                            */
/*============================================================================*/
/* Read hardware counters                                                     */

int
PCL_driver_read(
		PCL_DESCR_TYPE d, /* I: descriptor */
		int max_counter,
		PCL_BIT_MASK_TYPE counter_used_mask,
		PCL_CNT_TYPE *counter_values
		)

{
  int i, j;
  long long clocks;
  PCL_BIT_MASK_TYPE used_mask;
  PCL_state_descr *descr;
  HW_state_descr *hw_descr;


  /* get state descriptor */
  descr = d;
  hw_descr = &descr->hw_state_descr;

  /* set up used mask */
  used_mask = counter_used_mask & PCL_mask[max_counter-1];

  if(descr->event_table == event_table_pentium4)
    {
      /* get actual cycle count */
      if(IS_USED(used_mask, 18))
	{
	  rdtsc(clocks);
	  counter_values[18] = clocks - hw_descr->clock_value;
	}

      /* read out performance counters */
      i = 0;
      for(j = 0; j < 18; j++)
	{
	  if(IS_USED(used_mask, j))
	    {
	      counter_values[j] = vperfctr_read_pmc(hw_descr->self, i++);
	      //printf("counter %d: %lld\n", j, counter_values[j]);
	    }
	}
    }
  
  else if((descr->event_table == event_table_pentium)
	  || (descr->event_table == event_table_pentium_mmx)
	  || (descr->event_table == event_table_pentium2))
    {
      /* Pentium - Pentium III */
      /* get actual cycle count */
      if(IS_USED(used_mask, 2))
	{
	  rdtsc(clocks);
	  counter_values[2] = clocks - hw_descr->clock_value;
	}
      
      /* read out at most two performance counters */
      if(IS_USED(used_mask, 0))
	counter_values[0] = vperfctr_read_pmc(hw_descr->self, 0);
      if(IS_USED(used_mask, 1))
	counter_values[1] = vperfctr_read_pmc(hw_descr->self, 1);
    }
  
  else if(descr->event_table == event_table_athlon)
    {
      /* get actual cycle count */
      if(IS_USED(used_mask, 4))
	{
	  rdtsc(clocks);
	  counter_values[4] = clocks - hw_descr->clock_value;
	}
      
      /* read out at most four performance counters */
      if(IS_USED(used_mask, 0))
	counter_values[0] = vperfctr_read_pmc(hw_descr->self, 0);
      if(IS_USED(used_mask, 1))
	counter_values[1] = vperfctr_read_pmc(hw_descr->self, 1);
      if(IS_USED(used_mask, 2))
	counter_values[2] = vperfctr_read_pmc(hw_descr->self, 2);
      if(IS_USED(used_mask, 3))
	counter_values[3] = vperfctr_read_pmc(hw_descr->self, 3);
    }

  else
    /* should not happen (unknown event table) */
    return PCL_FAILURE;

  return PCL_SUCCESS;
}


/*                                                                            */
/*============================================================================*/
/*                                                                            */


int
PCL_driver_stop(
		PCL_DESCR_TYPE d /* I: descriptor */
		)

{
  PCL_state_descr *descr;
  HW_state_descr *hw_descr;


  /* get state descriptor */
  descr = d;
  hw_descr = &descr->hw_state_descr;

  /* stop counting */
  if(vperfctr_stop(hw_descr->self) == -1)
    return PCL_FAILURE;

  /* everything ok */
  return PCL_SUCCESS;
}


/*                                                                            */
/*============================================================================*/
/* close                                                                      */

int
PCL_driver_close(
		 PCL_DESCR_TYPE d /* I: descriptor */
		 )

{
  PCL_state_descr *descr;
  HW_state_descr *hw_descr;


  /* get state descriptor */
  descr = d;
  hw_descr = &descr->hw_state_descr;

  /* unlink driver (mmap etc.) */
  if(vperfctr_unlink(hw_descr->self) < 0)
    return PCL_FAILURE;

  /* close connection */
  vperfctr_close(hw_descr->self);

  /* everything ok */
  return PCL_SUCCESS;
}


/*============================================================================*
 *                             that's all folks                               *
 *============================================================================*/
