/*==============================================================================


   Purpose:    state descriptor


   Author:     Rudolf Berrendorf
               University of Applied Sciences Bonn-Rhein-Sieg
               Computer Science Department
               53754 Sankt Augustin
               rudolf.berrendorf@fh-bonn-rhein-sieg.de


   Date:       04.07.2000


   DISCLAIMER

                  PCL - The Performance Counter Library


   Copyright (C) 1999, 2000
   Forschungszentrum Juelich GmbH, Federal Republic of Germany.
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are met:

   Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

     - Redistributions of source code must retain the above copyright notice,
       this list of conditions and the following disclaimer.

     - Redistributions in binary form must reproduce the above copyright
       notice, this list of conditions and the following disclaimer in the
       documentation and/or other materials provided with the distribution.

     - Any publications that result from the use of this software shall
       reasonably refer to the Research Centre's development.

     - All advertising materials mentioning features or use of this software
       must display the following acknowledgement:

           This product includes software developed by Forschungszentrum
           Juelich GmbH, Federal Republic of Germany.

     - Forschungszentrum Juelich GmbH is not obligated to provide the user with
       any support, consulting, training or assistance of any kind with regard
       to the use, operation and performance of this software or to provide
       the user with any updates, revisions or new versions.


   THIS SOFTWARE IS PROVIDED BY FORSCHUNGSZENTRUM JUELICH GMBH "AS IS" AND ANY
   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
   DISCLAIMED. IN NO EVENT SHALL FORSCHUNGSZENTRUM JUELICH GMBH BE LIABLE FOR
   ANY SPECIAL, DIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
   RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
   CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
   CONNECTION WITH THE ACCESS, USE OR PERFORMANCE OF THIS SOFTWARE.


==============================================================================*/

#if !defined(STATE_DESCR_INCLUDED)
#define STATE_DESCR_INCLUDED

#include "pcl.h"


/*                                                                            */
/*============================================================================*/
/* Include hardware/os dependend libraries, global variables and definitions  */

/* For every processor, PCL_MAX_COUNTER must be defined which gives the maximum
   number of hardware performance counters. If you summarize several processors
   families with a different number of performance counters under one category,
   take the maximum value.  Some systems have a special cycle counter for
   elapsed cycles. You should include this counter in the definition of
   PCL_MAX_COUNTER.
*/


/*----------------------------------------------------------------------------*/
/* settings for Alpha_21164-Unicos/mk-SGI/CRAY. */

#if defined(ALPHA_CRAY)

#include "machines/ALPHA_CRAY/hw_driver_defs.h"

/*----------------------------------------------------------------------------*/
/* settings for  Alpha_21164/21264-True64-DEC/COMPAQ. */

#elif defined(ALPHA_DEC)

#include "machines/ALPHA_DEC/hw_driver_defs.h"

/*----------------------------------------------------------------------------*/
/* settings for MIPS_R1?000-IRIX-SGI. */

#elif defined(MIPS)

#include "machines/MIPS/hw_driver_defs.h"

/*----------------------------------------------------------------------------*/
/* settings for UltraSPARC-Solaris-SUN. */

#elif defined(SPARC)

#include "machines/SPARC/hw_driver_defs.h"

/*----------------------------------------------------------------------------*/
/* settings for *-AIX-IBM. */

#elif defined(POWERPC)

#include "machines/POWERPC/hw_driver_defs.h"

/*----------------------------------------------------------------------------*/
/* settings for Pentium/PPro/PII/PIII/P4/AMD-Linux-* */

#elif defined(PENTIUM_LINUX)

#include "machines/PENTIUM_LINUX/hw_driver_defs.h"

/*----------------------------------------------------------------------------*/
/* settings for IA64-Linux-* */

#elif defined(IA64_LINUX)

#include "machines/IA64_LINUX/hw_driver_defs.h"

/*----------------------------------------------------------------------------*/
/* settings for SR8000-* */

#elif defined(SR8000)

#include "machines/SR8000/hw_driver_defs.h"

/*----------------------------------------------------------------------------*/
/* else */

#else
#error "unsupported system"

#endif


#include "event_descr.h"


/*----------------------------------------------------------------------------*/
/* Inverse table entries. Given the results in ctr_0,...,ctr_n, this table
   specifies, how the PCL result values are constructed and where they are
   stored. Each hardware counter has such an entry.
*/

typedef struct
{
  PCL_CNT_TYPE result_value;	       /* result value of low-level driver */
  PCL_DRIVER_COMMAND_TYPE command;     /* command to activate low-level cntr */
  short result_index;		       /* contributes to what user result */
  short event;			       /* used on what event */
  unsigned short operation;	       /* combine operation */
}
PCL_result_descr_t;


typedef struct
{
  /* some magic value */
  int magic;

  /*--------------------------------------------------------------------------*/
  /* variables to generate result values. */

  /* performance counter values when starting nesting level */
  PCL_CNT_TYPE start_values[PCL_MAX_NESTING_LEVEL][PCL_MAX_COUNTER];

  /* sum of values gotton so far from performance counters at a specific
     nesting level. E.g. on interrupts the current values of the performance
     counters are read out and added to the values on the current nesting
     level. */
  PCL_CNT_TYPE virtual_counters[PCL_MAX_NESTING_LEVEL][PCL_MAX_COUNTER];
  PCL_BIT_MASK_TYPE virtual_counters_used_mask;


  /*--------------------------------------------------------------------------*/
  /* these variables store arguments to PCLread for consistency checking on
     subsequent PCLread/PCLstop calls. */

  /* count mode */
  unsigned int count_mode;

  /* number of requested events (must be same on each nesting level) */
  int event_count;

  /* requested events (must be same on each nesting level) */
  int events[PCL_MAX_EVENT_PER_CALL];


  /*--------------------------------------------------------------------------*/
  /* misc. variables */

  /* initialized? */
  int initialized;

  /* nesting level of PCL routines. Is incremented when PCLstart is called, and 
     decremented when PCLstop is called. */
  int nesting_level;

  /* bit mask which counters are needed */
  PCL_BIT_MASK_TYPE counter_used_mask;

  /* which events generate FP results */
  PCL_BIT_MASK_TYPE fp_result_mask;

  /* MHz-rate of processor */
  PCL_CNT_TYPE mhz_rate;

  /* for each low-level counter one entry */
  PCL_result_descr_t results[PCL_MAX_COUNTER];

#if defined(COUNTER_SIZE_32)	       /* 32-bit performance counters */
  /* old signal function */
  void (*old_signal_handler) (int);
#endif

  /*--------------------------------------------------------------------------*/
  /* these variables get initialized by the hardware driver */

  /* table of event descriptors */
  const PCL_event_descr_t *event_table;

  /* OR-ed bit mask of allowed count modes */
  unsigned int allowed_count_mode;

  /* hardware-dependend state descriptor */
  HW_state_descr hw_state_descr;

}
PCL_state_descr;



#endif				       /* !defined(STATE_DESCR_INCLUDED) */


/*============================================================================*
 *                             that's all folks                               *
 *============================================================================*/
